import csv
import re
from contextlib import contextmanager
from pathlib import Path
from tempfile import TemporaryDirectory

from snowfakery import SnowfakeryApplication, generate_data

from cumulusci.core.exceptions import BulkDataException, TaskOptionsError
from cumulusci.core.utils import (
    process_bool_arg,
    process_list_arg,
    process_list_of_pairs_dict_arg,
)
from cumulusci.tasks.bulkdata.step import (
    DataApi,
    DataOperationStatus,
    DataOperationType,
    get_dml_operation,
    get_query_operation,
)
from cumulusci.tasks.bulkdata.utils import RowErrorChecker
from cumulusci.tasks.salesforce import BaseSalesforceApiTask


class UpdateData(BaseSalesforceApiTask):
    """Update records of an sObject matching a where-clause."""

    task_options = {
        "object": {
            "description": "An SObject",
            "required": True,
        },
        "where": {
            "description": "A SOQL where-clause (without the keyword WHERE).",
            "required": False,
        },
        "api": {
            "description": "The desired Salesforce API to use, which may be 'rest', 'bulk', or "
            "'smart' to auto-select based on record volume. The default is 'smart'.",
            "required": False,
        },
        "fields": {
            "description": "Fields to download as input to the Snowfakery recipe",
            "required": False,
        },
        "recipe": {
            "description": "Snowfakery recipe to be executed on each row",
            "required": True,
        },
        "recipe_options": {
            "required": False,
            "description": """Pass values to override options in the format VAR1:foo,VAR2:bar

            Example: --recipe_options weight:10,color:purple""",
        },
        "ignore_row_errors": {
            "description": "If True, allow the operation to continue even if individual rows fail to delete."
        },
    }
    row_warning_limit = 10

    def _init_options(self, kwargs):
        super()._init_options(kwargs)

        self.sobject = self.options["object"]

        def identifier(f):
            return isinstance(f, str) and f.isidentifier()

        if not identifier(self.sobject):
            raise TaskOptionsError(
                f"`object` option should be an sObject name, not {self.sobject}"
            )

        self.where = self.options.get("where")
        self.fields = process_list_arg(self.options.get("fields", []))
        bad_fields = [f for f in self.fields if not identifier(f)]
        if bad_fields:
            plural = "s" if len(bad_fields) > 1 else ""
            verb = "are" if len(bad_fields) > 1 else "is"
            article = "an " if len(bad_fields) == 1 else ""
            raise TaskOptionsError(
                f"Field{plural} `{'`,`'.join(bad_fields)}` {verb} not {article}identifier{plural}"
            )

        self.inject_namespaces = process_bool_arg(
            self.task_options.get("inject_namespaces", True)
        )
        self.recipe = self.options.get("recipe")
        self.recipe_options = process_list_of_pairs_dict_arg(
            self.options.get("recipe_options", {})
        )
        self.ignore_row_errors = process_bool_arg(
            self.options.get("ignore_row_errors", False)
        )

        try:
            self.api = {
                "bulk": DataApi.BULK,
                "rest": DataApi.REST,
                "smart": DataApi.SMART,
            }[self.options.get("api", "smart").lower()]
        except KeyError:
            raise TaskOptionsError(
                f"{self.options['api']} is not a valid value for API (valid: bulk, rest, smart)"
            )

    def _run_task(self):
        obj = self._validate_and_inject_namespace_prefixes(
            should_inject_namespaces=self.inject_namespaces,
            sobjects_to_validate=[self.sobject],
            operation_to_validate="updateable",
        )[0]
        fields = ["Id"] + self.fields
        qs = self.query_objects(obj, fields)
        if not qs:
            self.logger.info(f"No records found, skipping update operation for {obj}")
            return
        with self.save_records(qs, fields) as csvfile, TemporaryDirectory() as outdir:
            csv_out = self.generate_data(csvfile, Path(outdir))

            with csv_out.open() as csv_out_open:
                enriched_data = csv.DictReader(csv_out_open)
                oid_index = enriched_data.fieldnames.index("Oid")
                enriched_data.fieldnames[oid_index] = "Id"
                ds = self.load_data(qs.job_result.records_processed, enriched_data)
            records = ds.job_result.records_processed
            errors = ds.job_result.total_row_errors

        obj_description = self._object_description(obj).capitalize()
        if errors:
            self.logger.info(
                f"{obj_description} processed ({records}). {errors} errors"
            )
        else:
            self.logger.info(f"{obj_description} successfully updated ({records}).")

        if errors and not self.ignore_row_errors:
            plural = "s" if errors > 1 else ""
            raise BulkDataException(f"{errors} update error{plural}")
        else:
            self.return_values = {**ds.job_result.simplify()}
        return self.return_values

    def generate_data(self, csvfile: Path, outdir: Path) -> Path:
        generate_data(
            yaml_file=self.recipe,
            user_options=self.recipe_options,
            update_input_file=csvfile,
            update_passthrough_fields=["Oid"],
            plugin_options={
                "org_config": self.org_config,
                "project_config": self.project_config,
            },
            output_folder=outdir,
            output_format="csv",
            parent_application=CumulusCIUpdatesApplication(self.logger),
        )
        created_csv = tuple(outdir.glob("*.csv"))
        assert len(created_csv) == 1, "CSV was not created by Snowfakery"
        return created_csv[0]

    @contextmanager
    def save_records(self, qs, fields):  # -> Path
        with TemporaryDirectory() as t:
            csvfile = Path(t) / "input.csv"
            with csvfile.open("w") as f:
                writer = csv.writer(f)
                snowfakery_fieldnames = ["Oid"] + fields[1:]
                writer.writerow(snowfakery_fieldnames)
                writer.writerows(qs.get_results())

            yield csvfile

    def query_objects(self, obj, fields):
        query = f"SELECT {','.join(fields)} FROM {obj}"
        if self.where:
            query += f" WHERE {self.options['where']}"

        qs = get_query_operation(
            sobject=obj,
            fields=fields,
            api_options={},
            context=self,
            query=query,
            api=DataApi.SMART,  # maybe this should always be bulk
        )

        self.logger.info(f"Querying for {self._object_description(obj)}")
        qs.query()
        if qs.job_result.status is not DataOperationStatus.SUCCESS:
            raise BulkDataException(
                f"Unable to query records for {obj}: {','.join(qs.job_result.job_errors)}"
            )
        if not qs.job_result.records_processed:
            return None
        return qs

    def load_data(self, row_count: int, records: csv.DictReader):
        obj = self.sobject
        self.logger.info(f"Updating {row_count} {obj} records")
        fieldnames = [
            f for f in records.fieldnames if f != "id" and not f.startswith("_")
        ]

        ds = get_dml_operation(
            sobject=obj,
            operation=DataOperationType.UPDATE,
            fields=fieldnames,
            api_options={},
            context=self,
            api=self.api,
            volume=row_count,
        )

        def cleanup(record):
            return tuple(record[fieldname] for fieldname in fieldnames)

        ds.start()
        ds.load_records(cleanup(record) for record in records)
        ds.end()

        if ds.job_result.status not in [
            DataOperationStatus.SUCCESS,
            DataOperationStatus.ROW_FAILURE,
        ]:
            raise BulkDataException(
                f"Unable to update records for {obj}: {','.join(ds.job_result.job_errors)}"
            )

        # "ignore" row errors because at this point it's too late. We're done the
        # job entirely. Raising an exception isn't very helpful. We can raise
        # it later.
        error_checker = RowErrorChecker(self.logger, True, self.row_warning_limit)
        for result in ds.get_results():
            error_checker.check_for_row_error(result, result.id)
        return ds

    def _object_description(self, obj):
        """Return a readable description of the object set to update."""
        if self.where:
            return f'{obj} objects matching "{self.options["where"]}"'
        else:
            return f"all {obj} objects"


class CumulusCIUpdatesApplication(SnowfakeryApplication):
    """Takes over Snowfakery logging so CumulusCI can control it"""

    MATCHER = re.compile(r"^Created [^ ]+.(csv|json)$")

    def __init__(self, logger) -> None:
        self.logger = logger
        super().__init__()

    def echo(self, message, *args, **kwargs):
        # skip CSV creation messages
        if not self.MATCHER.match(message):
            self.logger.info(message)



from unittest.mock import Mock
import base64
import io
import json
import os
import pathlib
import zipfile

import pytest

from cumulusci.tasks.salesforce.org_settings import (
    DeployOrgSettings,
    build_settings_package,
)
from cumulusci.utils import temporary_dir
from .util import create_task


class TestDeployOrgSettings:
    def test_run_task__json_only(self):
        with temporary_dir() as d:
            with open("dev.json", "w") as f:
                json.dump(
                    {
                        "settings": {
                            "orgPreferenceSettings": {"s1DesktopEnabled": True},
                            "otherSettings": {
                                "nested": {
                                    "boolValue": True,
                                    "stringValue": "string",
                                },
                            },
                        },
                    },
                    f,
                )
            path = os.path.join(d, "dev.json")
            task_options = {"definition_file": path, "api_version": "48.0"}
            task = create_task(DeployOrgSettings, task_options)
            task.api_class = Mock()
            task()

        package_zip = task.api_class.call_args[0][1]
        zf = zipfile.ZipFile(io.BytesIO(base64.b64decode(package_zip)), "r")
        assert (
            readtext(zf, "package.xml")
            == """<?xml version="1.0" encoding="UTF-8"?>
<Package xmlns="http://soap.sforce.com/2006/04/metadata">
    <types>
        <members>OrgPreference</members>
        <members>Other</members>
        <name>Settings</name>
    </types>
    <version>48.0</version>
</Package>"""
        )
        assert (
            readtext(zf, "settings/OrgPreference.settings")
            == """<?xml version="1.0" encoding="UTF-8"?>
<OrgPreferenceSettings xmlns="http://soap.sforce.com/2006/04/metadata">
    <preferences>
        <settingName>S1DesktopEnabled</settingName>
        <settingValue>True</settingValue>
    </preferences>
</OrgPreferenceSettings>"""
        )
        assert (
            readtext(zf, "settings/Other.settings")
            == """<?xml version="1.0" encoding="UTF-8"?>
<OtherSettings xmlns="http://soap.sforce.com/2006/04/metadata">
    <nested>
        <boolValue>true</boolValue>
        <stringValue>string</stringValue>
    </nested>
</OtherSettings>"""
        )

    def test_run_task__settings_only(self):
        settings = {
            "orgPreferenceSettings": {"s1DesktopEnabled": True},
            "otherSettings": {
                "nested": {
                    "boolValue": True,
                    "stringValue": "string",
                },
            },
        }
        task_options = {"settings": settings, "api_version": "48.0"}
        task = create_task(DeployOrgSettings, task_options)
        task.api_class = Mock()
        task()

        package_zip = task.api_class.call_args[0][1]
        zf = zipfile.ZipFile(io.BytesIO(base64.b64decode(package_zip)), "r")
        assert (
            readtext(zf, "package.xml")
            == """<?xml version="1.0" encoding="UTF-8"?>
<Package xmlns="http://soap.sforce.com/2006/04/metadata">
    <types>
        <members>OrgPreference</members>
        <members>Other</members>
        <name>Settings</name>
    </types>
    <version>48.0</version>
</Package>"""
        )
        assert (
            readtext(zf, "settings/OrgPreference.settings")
            == """<?xml version="1.0" encoding="UTF-8"?>
<OrgPreferenceSettings xmlns="http://soap.sforce.com/2006/04/metadata">
    <preferences>
        <settingName>S1DesktopEnabled</settingName>
        <settingValue>True</settingValue>
    </preferences>
</OrgPreferenceSettings>"""
        )
        assert (
            readtext(zf, "settings/Other.settings")
            == """<?xml version="1.0" encoding="UTF-8"?>
<OtherSettings xmlns="http://soap.sforce.com/2006/04/metadata">
    <nested>
        <boolValue>true</boolValue>
        <stringValue>string</stringValue>
    </nested>
</OtherSettings>"""
        )

    def test_run_task__json_and_settings(self):
        with temporary_dir() as d:
            with open("dev.json", "w") as f:
                json.dump(
                    {
                        "settings": {
                            "orgPreferenceSettings": {"s1DesktopEnabled": True},
                        },
                    },
                    f,
                )
            path = os.path.join(d, "dev.json")
            settings = {
                "otherSettings": {
                    "nested": {
                        "boolValue": True,
                        "stringValue": "string",
                    },
                },
            }
            task_options = {
                "definition_file": path,
                "settings": settings,
                "api_version": "48.0",
            }
            task = create_task(DeployOrgSettings, task_options)
            task.api_class = Mock()
            task()

        package_zip = task.api_class.call_args[0][1]
        zf = zipfile.ZipFile(io.BytesIO(base64.b64decode(package_zip)), "r")
        assert (
            readtext(zf, "package.xml")
            == """<?xml version="1.0" encoding="UTF-8"?>
<Package xmlns="http://soap.sforce.com/2006/04/metadata">
    <types>
        <members>OrgPreference</members>
        <members>Other</members>
        <name>Settings</name>
    </types>
    <version>48.0</version>
</Package>"""
        )
        assert (
            readtext(zf, "settings/OrgPreference.settings")
            == """<?xml version="1.0" encoding="UTF-8"?>
<OrgPreferenceSettings xmlns="http://soap.sforce.com/2006/04/metadata">
    <preferences>
        <settingName>S1DesktopEnabled</settingName>
        <settingValue>True</settingValue>
    </preferences>
</OrgPreferenceSettings>"""
        )
        assert (
            readtext(zf, "settings/Other.settings")
            == """<?xml version="1.0" encoding="UTF-8"?>
<OtherSettings xmlns="http://soap.sforce.com/2006/04/metadata">
    <nested>
        <boolValue>true</boolValue>
        <stringValue>string</stringValue>
    </nested>
</OtherSettings>"""
        )

    def test_run_task__no_settings(self):
        with temporary_dir() as d:
            with open("dev.json", "w") as f:
                f.write("{}")
            path = os.path.join(d, "dev.json")
            task_options = {"definition_file": path}
            task = create_task(DeployOrgSettings, task_options)
            task.api_class = Mock()
            task()
            assert not task.api_class.called

    def test_run_task_bad_settings_type(self):
        with temporary_dir() as d:
            with open("dev.json", "w") as f:
                json.dump(
                    {
                        "settings": {
                            "otherSettings": {
                                "list": [],
                            },
                        },
                    },
                    f,
                )
            path = os.path.join(d, "dev.json")
            task_options = {"definition_file": path, "api_version": "48.0"}
            task = create_task(DeployOrgSettings, task_options)
            task.api_class = Mock()
            with pytest.raises(TypeError):
                task()


def readtext(zf, name):
    with zf.open(name, "r") as f:
        return io.TextIOWrapper(f).read()


class TestBuildSettingsPackage:
    def test_build_settings_package(self):
        settings = {
            "otherSettings": {
                "nested": {
                    "boolValue": True,
                    "stringValue": "string",
                },
            }
        }
        object_settings = {
            "account": {
                "defaultRecordType": "Default",
                "sharingModel": "Public",
            },
            "solution": {"defaultRecordType": "Default"},
        }
        with build_settings_package(settings, object_settings, "48.0") as path:
            assert (
                (pathlib.Path(path) / "package.xml").read_text()
                == """<?xml version="1.0" encoding="UTF-8"?>
<Package xmlns="http://soap.sforce.com/2006/04/metadata">
    <types>
        <members>Solution.DefaultSolution</members>
        <name>BusinessProcess</name>
    </types>
    <types>
        <members>Account.Default</members>
        <members>Solution.Default</members>
        <name>RecordType</name>
    </types>
    <types>
        <members>Other</members>
        <name>Settings</name>
    </types>
    <version>48.0</version>
</Package>"""
            )
            assert (
                (pathlib.Path(path) / "settings" / "Other.settings").read_text()
                == """<?xml version="1.0" encoding="UTF-8"?>
<OtherSettings xmlns="http://soap.sforce.com/2006/04/metadata">
    <nested>
        <boolValue>true</boolValue>
        <stringValue>string</stringValue>
    </nested>
</OtherSettings>"""
            )
            assert (
                (pathlib.Path(path) / "objects" / "Account.object").read_text()
                == """<?xml version="1.0" encoding="UTF-8"?>
<Object xmlns="http://soap.sforce.com/2006/04/metadata">
    <sharingModel>Public</sharingModel>
    <recordTypes>
        <fullName>Default</fullName>
        <label>Default</label>
        <active>true</active>
        
    </recordTypes>
        

</Object>"""
            )
            print((pathlib.Path(path) / "objects" / "Solution.object").read_text())
            assert (
                (pathlib.Path(path) / "objects" / "Solution.object").read_text()
                == """<?xml version="1.0" encoding="UTF-8"?>
<Object xmlns="http://soap.sforce.com/2006/04/metadata">

    <recordTypes>
        <fullName>Default</fullName>
        <label>Default</label>
        <active>true</active>
        <businessProcess>DefaultSolution</businessProcess>
    </recordTypes>
        
    <businessProcesses>
        <fullName>DefaultSolution</fullName>
        <isActive>true</isActive>
        <values>
            <fullName>Draft</fullName>
            
        </values>
    </businessProcesses>
</Object>"""
            )



from datetime import date
from enum import Enum
from logging import getLogger
from pathlib import Path
from typing import IO, Any, Callable, Dict, List, Mapping, Optional, Union

from pydantic import Field, ValidationError, root_validator, validator
from requests.structures import CaseInsensitiveDict as RequestsCaseInsensitiveDict
from simple_salesforce import Salesforce
from typing_extensions import Literal

from cumulusci.core.exceptions import BulkDataException
from cumulusci.tasks.bulkdata.dates import iso_to_date
from cumulusci.tasks.bulkdata.step import DataApi, DataOperationType
from cumulusci.utils import convert_to_snake_case
from cumulusci.utils.yaml.model_parser import CCIDictModel

logger = getLogger(__name__)


class CaseInsensitiveDict(RequestsCaseInsensitiveDict):
    def __init__(self, *args, **kwargs):
        self._canonical_keys = {}
        super().__init__(*args, **kwargs)

    def canonical_key(self, name):
        return self._canonical_keys[name.lower()]

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self._canonical_keys[key.lower()] = key


class MappingLookup(CCIDictModel):
    "Lookup relationship between two tables."
    table: str
    key_field: Optional[str] = None
    value_field: Optional[str] = None
    join_field: Optional[str] = None
    after: Optional[str] = None
    aliased_table: Optional[Any] = None
    name: Optional[str] = None  # populated by parent

    def get_lookup_key_field(self, model=None):
        "Find the field name for this lookup."
        guesses = []
        if self.get("key_field"):
            guesses.append(self.get("key_field"))

        guesses.append(self.name)

        if not model:
            return guesses[0]

        # CCI used snake_case until mid-2020.
        # At some point this code could probably be simplified.
        snake_cased_guesses = list(map(convert_to_snake_case, guesses))
        guesses = guesses + snake_cased_guesses
        for guess in guesses:
            if hasattr(model, guess):
                return guess
        raise KeyError(
            f"Could not find a key field for {self.name}.\n"
            + f"Tried {', '.join(guesses)}"
        )


SHOULD_REPORT_RECORD_TYPE_DEPRECATION = True


class BulkMode(Enum):
    serial = Serial = "Serial"
    parallel = Parallel = "Parallel"


ENUM_VALUES = {
    v.value.lower(): v.value
    for enum in [BulkMode, DataApi, DataOperationType]
    for v in enum.__members__.values()
}


class MappingStep(CCIDictModel):
    "Step in a load or extract process"
    sf_object: str
    table: Optional[str] = None
    fields_: Dict[str, str] = Field({}, alias="fields")
    lookups: Dict[str, MappingLookup] = {}
    static: Dict[str, str] = {}
    filters: List[str] = []
    action: DataOperationType = DataOperationType.INSERT
    api: DataApi = DataApi.SMART
    batch_size: int = None
    oid_as_pk: bool = False  # this one should be discussed and probably deprecated
    record_type: Optional[str] = None  # should be discussed and probably deprecated
    bulk_mode: Optional[
        Literal["Serial", "Parallel"]
    ] = None  # default should come from task options
    anchor_date: Optional[Union[str, date]] = None
    soql_filter: Optional[str] = None  # soql_filter property

    @validator("bulk_mode", "api", "action", pre=True)
    def case_normalize(cls, val):
        if isinstance(val, Enum):
            return val
        return ENUM_VALUES.get(val.lower())

    def get_oid_as_pk(self):
        """Returns True if using Salesforce Ids as primary keys."""
        return "Id" in self.fields

    def get_destination_record_type_table(self):
        """Returns the name of the record type table for the target org."""
        return f"{self.sf_object}_rt_target_mapping"

    def get_source_record_type_table(self):
        """Returns the name of the record type table for the source org."""
        return f"{self.sf_object}_rt_mapping"

    def get_sf_id_table(self):
        """Returns the name of the table for storing Salesforce Ids."""
        return f"{self.table}_sf_ids"

    def get_complete_field_map(self, include_id=False):
        """Return a field map that includes both `fields` and `lookups`.
        If include_id is True, add the Id field if not already present."""
        fields = {}

        if include_id and "Id" not in self.fields:
            fields["Id"] = "sf_id"

        fields.update(self.fields)
        fields.update(
            {
                lookup: self.lookups[lookup].get_lookup_key_field()
                for lookup in self.lookups
            }
        )

        return fields

    def get_fields_by_type(self, field_type: str, sf: Salesforce):
        describe = getattr(sf, self.sf_object).describe()
        describe = CaseInsensitiveDict(
            {entry["name"]: entry for entry in describe["fields"]}
        )

        return [f for f in describe if describe[f]["type"] == field_type]

    def get_load_field_list(self):
        """Build a flat list of columns for the given mapping,
        including fields, lookups, and statics."""
        lookups = self.lookups

        # Build the list of fields to import
        columns = []
        columns.extend(self.fields.keys())

        # Don't include lookups with an `after:` spec (dependent lookups)
        columns.extend([f for f in lookups if not lookups[f].after])
        columns.extend(self.static.keys())

        # If we're using Record Type mapping, `RecordTypeId` goes at the end.
        if "RecordTypeId" in columns:
            columns.remove("RecordTypeId")

        if self.action is DataOperationType.INSERT and "Id" in columns:
            columns.remove("Id")
        if self.record_type or "RecordTypeId" in self.fields:
            columns.append("RecordTypeId")

        return columns

    def get_relative_date_context(self, fields: List[str], sf: Salesforce):
        date_fields = [
            fields.index(f)
            for f in self.get_fields_by_type("date", sf)
            if f in self.fields
        ]
        date_time_fields = [
            fields.index(f)
            for f in self.get_fields_by_type("datetime", sf)
            if f in self.fields
        ]

        return (date_fields, date_time_fields, date.today())

    @validator("batch_size")
    @classmethod
    def validate_batch_size(cls, v, values):
        if values["api"] == DataApi.REST:
            assert 0 < v <= 200, "Max 200 batch_size for REST loads"
        elif values["api"] == DataApi.BULK:
            assert 0 < v <= 10_000, "Max 10,000 batch_size for bulk or smart loads"
        elif values["api"] == DataApi.SMART and v is not None:
            assert 0 < v < 200, "Max 200 batch_size for Smart loads"
            logger.warning(
                "If you set a `batch_size` you should also set an `api` to `rest` or `bulk`. "
                "`batch_size` means different things for `rest` and `bulk`. "
                "Please see the documentation for further details. "
                "https://cumulusci.readthedocs.io/en/latest/data.html#api-selection"
            )
        else:  # pragma: no cover
            # should not happen
            assert f"Unknown API {values['api']}"
        return v

    @validator("anchor_date")
    @classmethod
    def validate_anchor_date(cls, v):
        return iso_to_date(v)

    @validator("record_type")
    @classmethod
    def record_type_is_deprecated(cls, v):
        if SHOULD_REPORT_RECORD_TYPE_DEPRECATION:
            logger.warning(
                "record_type is deprecated. Just supply a RecordTypeId column declaration and it will be inferred"
            )
        return v

    @validator("oid_as_pk")
    @classmethod
    def oid_as_pk_is_deprecated(cls, v):
        raise ValueError(
            "oid_as_pk is no longer supported. Include the Id field if desired."
        )

    @validator("fields_", pre=True)
    @classmethod
    def standardize_fields_to_dict(cls, values):
        if values is None:
            values = {}
        if type(values) is list:
            return {elem: elem for elem in values}

        return values

    @root_validator
    @classmethod
    def set_default_table(cls, values):
        """Automatically populate the `table` key with `sf_object`, if not present."""
        if values["table"] is None:
            values["table"] = values.get("sf_object")

        return values

    @root_validator  # not really a validator, more like a post-processor
    @classmethod
    def fixup_lookup_names(cls, v):
        "Allow lookup objects to know the key they were attached to in the mapping file."
        for name, lookup in v["lookups"].items():
            lookup.name = name
        return v

    @staticmethod
    def _is_injectable(element: str) -> bool:
        return element.count("__") == 1

    def _get_permission_type(self, operation: DataOperationType) -> str:
        if operation is DataOperationType.QUERY:
            return "queryable"
        if (
            operation is DataOperationType.INSERT
            and self.action is DataOperationType.UPDATE
        ):
            return "updateable"

        return "createable"

    def _check_object_permission(
        self, global_describe: Mapping, sobject: str, operation: DataOperationType
    ):
        assert sobject in global_describe
        perm = self._get_permission_type(operation)
        return global_describe[sobject][perm]

    def _check_field_permission(
        self, describe: Mapping, field: str, operation: DataOperationType
    ):
        perm = self._get_permission_type(operation)
        # Fields don't have "queryable" permission.
        return field in describe and (
            describe[field].get(perm) if perm in describe[field] else True
        )

    def _validate_field_dict(
        self,
        describe: CaseInsensitiveDict,
        field_dict: Dict[str, Any],
        inject: Optional[Callable[[str], str]],
        strip: Optional[Callable[[str], str]],
        drop_missing: bool,
        data_operation_type: DataOperationType,
    ) -> bool:
        ret = True

        def replace_if_necessary(dct, name, replacement):
            if name not in describe and replacement in describe:
                dct[replacement] = dct[name]
                del dct[name]
                return replacement
            else:
                return name

        orig_fields = field_dict.copy()
        for f, entry in orig_fields.items():
            # Do we need to inject this field?
            if f.lower() == "id":
                del field_dict[f]
                field_dict["Id"] = entry
                continue

            if inject and self._is_injectable(f) and inject(f) not in orig_fields:
                if f in describe and inject(f) in describe:
                    logger.warning(
                        f"Both {self.sf_object}.{f} and {self.sf_object}.{inject(f)} are present in the target org. Using {f}."
                    )

                f = replace_if_necessary(field_dict, f, inject(f))
            if strip:
                f = replace_if_necessary(field_dict, f, strip(f))

            # Canonicalize the key's case
            try:
                new_name = describe.canonical_key(f)
            except KeyError:
                logger.warning(
                    f"Field {self.sf_object}.{f} does not exist or is not visible to the current user."
                )
            else:
                del field_dict[f]
                field_dict[new_name] = entry
                f = new_name

            # Do we have the right permissions for this field, or do we need to drop it?
            is_after_lookup = hasattr(field_dict[f], "after")
            if not self._check_field_permission(
                describe,
                f,
                data_operation_type
                if not is_after_lookup
                else DataOperationType.UPDATE,
            ):
                logger.warning(
                    f"Field {self.sf_object}.{f} is not present or does not have the correct permissions."
                )
                if drop_missing:
                    del field_dict[f]
                else:
                    ret = False

        return ret

    def _validate_sobject(
        self,
        global_describe: CaseInsensitiveDict,
        inject: Optional[Callable[[str], str]],
        strip: Optional[Callable[[str], str]],
        data_operation_type: DataOperationType,
    ) -> bool:
        # Determine whether we need to inject or strip our sObject.

        self.sf_object = (
            _inject_or_strip_name(self.sf_object, inject, global_describe)
            or _inject_or_strip_name(self.sf_object, strip, global_describe)
            or self.sf_object
        )

        try:
            self.sf_object = global_describe.canonical_key(self.sf_object)
        except KeyError:
            logger.warning(
                f"sObject {self.sf_object} does not exist or is not visible to the current user."
            )
            return False

        # Validate our access to this sObject.
        if not self._check_object_permission(
            global_describe, self.sf_object, data_operation_type
        ):
            logger.warning(
                f"sObject {self.sf_object} does not have the correct permissions for {data_operation_type}."
            )
            return False

        return True

    def validate_and_inject_namespace(
        self,
        sf: Salesforce,
        namespace: Optional[str],
        operation: DataOperationType,
        inject_namespaces: bool = False,
        drop_missing: bool = False,
    ):
        """Process the schema elements in this step.

        First, we inject the namespace into object and field names where applicable.
        Second, we validate that all fields are accessible by the running user with the
        correct permission level for this operation.
        Lastly, if drop_missing is True, we strip any fields that are not present (namespaced
        or otherwise) from the target org.

        Return True if this object should be processed. If drop_missing is True, a False return
        value indicates we should skip this object. If drop_missing is False, a False return
        value indicates that one or more schema elements couldn't be validated."""

        if namespace and inject_namespaces:

            def inject(element: str):
                return f"{namespace}__{element}"

            def strip(element: str):
                parts = element.split("__")
                if len(parts) == 3 and parts[0] == namespace:
                    return parts[1] + "__" + parts[2]
                else:
                    return element

        else:
            inject = strip = None

        global_describe = CaseInsensitiveDict(
            {entry["name"]: entry for entry in sf.describe()["sobjects"]}
        )

        if not self._validate_sobject(global_describe, inject, strip, operation):
            # Don't attempt to validate field permissions if the object doesn't exist.
            return False

        # Validate, inject, and drop (if configured) fields.
        # By this point, we know the attribute is valid.
        describe = getattr(sf, self.sf_object).describe()
        describe = CaseInsensitiveDict(
            {entry["name"]: entry for entry in describe["fields"]}
        )

        if not self._validate_field_dict(
            describe, self.fields, inject, strip, drop_missing, operation
        ):
            return False

        if not self._validate_field_dict(
            describe, self.lookups, inject, strip, drop_missing, operation
        ):
            return False

        return True


class MappingSteps(CCIDictModel):
    "Mapping of named steps"
    __root__: Dict[str, MappingStep]

    @root_validator(pre=False)
    @classmethod
    def validate_and_inject_mapping(cls, values):
        if values:
            oids = ["Id" in s.fields_ for s in values["__root__"].values()]
            assert all(oids) or not any(
                oids
            ), "Id must be mapped in all steps or in no steps."

        return values


ValidationError = ValidationError  # export Pydantic's Validation Error under an alias


def parse_from_yaml(source: Union[str, Path, IO]) -> Dict:
    "Parse from a path, url, path-like or file-like"
    return MappingSteps.parse_from_yaml(source)


def validate_and_inject_mapping(
    *,
    mapping: Dict,
    sf: Salesforce,
    namespace: str,
    data_operation: DataOperationType,
    inject_namespaces: bool,
    drop_missing: bool,
    org_has_person_accounts_enabled: bool = False,
):
    should_continue = [
        m.validate_and_inject_namespace(
            sf, namespace, data_operation, inject_namespaces, drop_missing
        )
        for m in mapping.values()
    ]

    if not drop_missing and not all(should_continue):
        raise BulkDataException(
            "One or more schema or permissions errors blocked the operation.\n"
            "If you would like to attempt the load regardless, you can specify "
            "'-o drop_missing_schema True' on the command."
        )

    if drop_missing:
        # Drop any steps with sObjects that are not present.
        for (include, step_name) in zip(should_continue, list(mapping.keys())):
            if not include:
                del mapping[step_name]

        # Remove any remaining lookups to dropped objects.
        for m in mapping.values():
            describe = getattr(sf, m.sf_object).describe()
            describe = {entry["name"]: entry for entry in describe["fields"]}

            for field in list(m.lookups.keys()):
                lookup = m.lookups[field]
                if lookup.table not in [step.table for step in mapping.values()]:
                    del m.lookups[field]

                    # Make sure this didn't cause the operation to be invalid
                    # by dropping a required field.
                    if not describe[field]["nillable"]:
                        raise BulkDataException(
                            f"{m.sf_object}.{field} is a required field, but the target object "
                            f"{describe[field]['referenceTo']} was removed from the operation "
                            "due to missing permissions."
                        )

    # If the org has person accounts enable, add a field mapping to track "IsPersonAccount".
    # IsPersonAccount field values are used to properly load person account records.
    if org_has_person_accounts_enabled and data_operation == DataOperationType.QUERY:
        for step in mapping.values():
            if step["sf_object"] in ("Account", "Contact"):
                step["fields"]["IsPersonAccount"] = "IsPersonAccount"


def _inject_or_strip_name(name, transform, global_describe):
    if not transform:
        return None
    new_name = transform(name)

    if name == new_name:
        return None

    if name in global_describe and new_name in global_describe:
        logger.warning(
            f"Both {name} and {new_name} are present in the target org. Using {name}."
        )
        return None

    if name not in global_describe and new_name in global_describe:
        return new_name

    return None



from glob import glob
from pathlib import Path
from shutil import rmtree
from tempfile import TemporaryDirectory, mkdtemp

import pytest
from lxml import etree

from cumulusci.utils.xml import lxml_parse_file, lxml_parse_string
from cumulusci.utils.xml.salesforce_encoding import serialize_xml_for_salesforce


class TestSalesforceEncoding:
    def test_xml_declaration(self):
        xml = lxml_parse_string("<foo/>")
        out = serialize_xml_for_salesforce(xml, xml_declaration=True)
        assert out.startswith("<?xml")

        out = serialize_xml_for_salesforce(xml, xml_declaration=False)
        assert not out.startswith("<?xml")

        out = serialize_xml_for_salesforce(xml)
        assert out.startswith("<?xml")

    def test_roundtripping(self):
        from cumulusci.tasks.metadata import tests

        files = glob(str(Path(tests.__file__).parent / "sample_package.xml"))

        #   If you fiddle with the salesforce encoder, the code below may be
        #   useful to ensure that it faithfully round-trips, but it only works
        #   if run in a directory with a parent-directory which contains a few
        #   CCI projects which contain translations in them. At the time of
        #   writing it roundtripped 421 files without a byte of difference.
        #
        # files += glob("../**/*ranslations/**ranslation", recursive=True)
        # files += glob("../**/*-meta.xml", recursive=True)
        # files += glob("../**/package.xml", recursive=True)
        # files += glob("../**/*.object.xml", recursive=True)

        assert files

        problems = []

        temp_directory = mkdtemp(prefix="xmltests")
        for file in files:
            with open(file) as f:
                orig = f.read().strip()

            tree = lxml_parse_file(file)
            out = serialize_xml_for_salesforce(tree).strip()
            try:
                orig = etree.canonicalize(orig)
                out = etree.canonicalize(out)
                c19n_succeeded = True
            except Exception:  # pragma: no cover
                c19n_succeeded = False
            try:
                assert orig == out, file
            except Exception as e:  # pragma: no cover
                details = self._save_exception_for_inspection(
                    file, temp_directory, orig, out, c19n_succeeded, e
                )
                problems.append(details)

        assert not problems, (temp_directory, len(problems))
        rmtree(temp_directory)  # clean up if there was no assertion failure

    def _save_exception_for_inspection(
        self, file, temp_directory, orig, out, c19n_succeeded, exception
    ):  # pragma: no cover
        filename = Path(file).name
        infile_copy = str(Path(temp_directory) / filename)
        with open(infile_copy, "w") as f:
            f.write(orig)
        outfile_copy = infile_copy + ".out.xml"
        with open(outfile_copy, "w") as f:
            f.write(out)
        outfile_exception = infile_copy + ".err"
        with open(outfile_exception, "w") as f:
            f.write(file)
            f.write(f"C19 Status: {c19n_succeeded}")
            f.write("\n")
            f.write(str(exception))
        return (infile_copy, outfile_copy, outfile_exception, file)

    def tests_filename_bad_xml(self):
        with TemporaryDirectory() as d:
            filename = d + "/foobar.notxml"
            with open(filename, "w") as f:
                f.write("<<<<")
            with pytest.raises(SyntaxError) as e:
                lxml_parse_file(filename)
            assert "foobar.notxml" in str(e.value)

    def test_comment(self):
        tree = lxml_parse_string(
            "<Foo><!-- Salesforce files should not have comments, but just in case! --> </Foo>"
        )
        assert "just in case! -->" in serialize_xml_for_salesforce(tree)

    def test_empty_elements(self):
        xml_in = """
<Layout xmlns="http://soap.sforce.com/2006/04/metadata">
    <layoutSections>
        <layoutColumns/>
        <layoutColumns/>
        <layoutColumns/>
        <style>CustomLinks</style>
    </layoutSections>
</Layout>""".strip()

        tree = lxml_parse_string(xml_in)

        xml_out = serialize_xml_for_salesforce(tree, xml_declaration=False)
        assert xml_in == xml_out.strip()

    def test_namespaces(self):
        xml_in = '<Foo xmlns:foo="https://html5zombo.com/" xmlns:dad="http://niceonedad.com/"/>\n'
        tree = lxml_parse_string(xml_in)

        xml_out = serialize_xml_for_salesforce(tree, xml_declaration=False)
        assert xml_in == xml_out

    def test_simple_attributes_roundtrip(self):
        xml_in = """
<Layout xmlns="http://soap.sforce.com/2006/04/metadata">
    <layoutSections>
        <layoutColumns foo="bar"/>
    </layoutSections>
</Layout>""".strip()

        tree = lxml_parse_string(xml_in)

        xml_out = serialize_xml_for_salesforce(tree, xml_declaration=False)
        assert xml_in == xml_out.strip()

    def test_namespaced_attributes_roundtrip(self):
        xml_in = """
<CustomMetadata xmlns="http://soap.sforce.com/2006/04/metadata" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
    <label>Account MD Isolation Rollup</label>
    <protected>false</protected>
    <values>
        <field>dlrs__Active__c</field>
        <value xsi:type="xsd:boolean">true</value>
    </values>
</CustomMetadata>""".strip()

        tree = lxml_parse_string(xml_in)

        xml_out = serialize_xml_for_salesforce(tree, xml_declaration=False)
        assert xml_in == xml_out.strip()

    def test_with_element_rather_than_doc(self):
        xml_in = """
<CustomMetadata xmlns="http://soap.sforce.com/2006/04/metadata" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
    <label>Account MD Isolation Rollup</label>
    <protected>false</protected>
    <values>
        <field>dlrs__Active__c</field>
        <value xsi:type="xsd:boolean">true</value>
    </values>
</CustomMetadata>""".strip()

        tree = lxml_parse_string(xml_in)

        xml_out = serialize_xml_for_salesforce(tree.getroot(), xml_declaration=False)
        assert xml_in == xml_out.strip()

        xml_out = serialize_xml_for_salesforce(
            tree.getroot()[-1][0], xml_declaration=False
        )
        assert xml_out.strip() == "<field>dlrs__Active__c</field>"



from contextlib import contextmanager

import pytest

from cumulusci.tasks.bulkdata.delete import DeleteData


@pytest.fixture()
def ensure_accounts(create_task, run_code_without_recording, sf):
    """Delete all accounts and create a certain number of new ones"""

    @contextmanager
    def _ensure_accounts(number_of_accounts):
        def setup(number):
            task = create_task(DeleteData, {"objects": "Entitlement, Account"})
            task()
            for i in range(0, number):
                sf.Account.create({"Name": f"Account {i}"})

        run_code_without_recording(lambda: setup(number_of_accounts))
        yield
        run_code_without_recording(lambda: setup(0))

    return _ensure_accounts



import io
from unittest import mock
from zipfile import ZipFile

import pytest

from cumulusci.core.exceptions import SfdxOrgException
from cumulusci.core.sfdx import (
    SourceFormat,
    convert_sfdx_source,
    get_default_devhub_username,
    get_source_format_for_path,
    get_source_format_for_zipfile,
    sfdx,
    shell_quote,
)
from cumulusci.utils import temporary_dir, touch


class TestSfdx:
    @mock.patch("platform.system", mock.Mock(return_value="Linux"))
    @mock.patch("sarge.Command")
    def test_posix_quoting(self, Command):
        sfdx("cmd", args=["a'b"])
        cmd = Command.call_args[0][0]
        assert cmd == r"sfdx cmd 'a'\''b'"

    @mock.patch("platform.system", mock.Mock(return_value="Windows"))
    @mock.patch("sarge.Command")
    def test_windows_quoting(self, Command):
        sfdx("cmd", args=['a"b'], access_token="token")
        cmd = Command.call_args[0][0]
        assert cmd == r'sfdx cmd "a\"b" -u token'

    @mock.patch("platform.system", mock.Mock(return_value="Windows"))
    def test_shell_quote__str_with_space(self):
        actual = shell_quote("pkg-name Managed Feature Test")
        assert '"pkg-name Managed Feature Test"' == actual

    @mock.patch("sarge.Command")
    def test_check_return(self, Command):
        Command.return_value.returncode = 1
        Command.return_value.stderr = io.BytesIO(b"Egads!")
        with pytest.raises(Exception) as exc_info:
            sfdx("cmd", check_return=True)
        assert str(exc_info.value) == "Command exited with return code 1:\nEgads!"


@mock.patch("sarge.Command")
def test_get_default_devhub_username(Command):
    Command.return_value.returncode = 0
    Command.return_value.stdout = io.BytesIO(
        b'{"result": [{"value": "devhub@example.com"}]}'
    )
    result = get_default_devhub_username()
    assert result == "devhub@example.com"


@mock.patch("sarge.Command")
def test_get_default_devhub_username__no_result(Command):
    Command.return_value.returncode = 0
    Command.return_value.stdout = io.BytesIO(b"{}")
    with pytest.raises(SfdxOrgException):
        get_default_devhub_username()


def test_get_source_format_for_path():
    with temporary_dir(chdir=True) as td:
        assert get_source_format_for_path(td) is SourceFormat.SFDX

        touch("package.xml")
        assert get_source_format_for_path(td) is SourceFormat.MDAPI


def test_get_source_format_for_zipfile():
    zf = ZipFile(io.BytesIO(), "w")

    zf.writestr("package.xml", "test")
    zf.writestr("src/package.xml", "test")
    zf.writestr("force-app/foo.txt", "test")

    assert get_source_format_for_zipfile(zf, None) is SourceFormat.MDAPI
    assert get_source_format_for_zipfile(zf, "src") is SourceFormat.MDAPI
    assert get_source_format_for_zipfile(zf, "force-app") is SourceFormat.SFDX


def test_convert_sfdx():
    logger = mock.Mock()
    with temporary_dir() as path:
        touch("README.md")  # make sure there's something in the directory
        with mock.patch("cumulusci.core.sfdx.sfdx") as sfdx:
            with convert_sfdx_source(path, "Test Package", logger) as p:
                assert p is not None

    sfdx.assert_called_once_with(
        "force:source:convert",
        args=["-d", mock.ANY, "-r", path, "-n", "Test Package"],
        capture_output=True,
        check_return=True,
    )


def test_convert_sfdx__cwd():
    logger = mock.Mock()
    with temporary_dir(chdir=True):
        touch("README.md")  # make sure there's something in the directory
        with mock.patch("cumulusci.core.sfdx.sfdx") as sfdx:
            with convert_sfdx_source(None, "Test Package", logger) as p:
                assert p is not None

    sfdx.assert_called_once_with(
        "force:source:convert",
        args=["-d", mock.ANY, "-n", "Test Package"],
        capture_output=True,
        check_return=True,
    )


def test_convert_sfdx__mdapi():
    logger = mock.Mock()
    with temporary_dir() as path:
        touch("package.xml")  # make sure there's something in the directory
        with mock.patch("cumulusci.core.sfdx.sfdx") as sfdx:
            with convert_sfdx_source(path, "Test Package", logger) as p:
                assert p == path

    sfdx.assert_not_called()


def test_convert_sfdx__skipped_if_directory_empty():
    logger = mock.Mock()
    with temporary_dir() as path:
        with mock.patch("cumulusci.core.sfdx.sfdx") as sfdx:
            with convert_sfdx_source(path, "Test Package", logger):
                pass

    sfdx.assert_not_called()



import datetime
import json
import os
import re

from cumulusci.core.config import FAILED_TO_CREATE_SCRATCH_ORG
from cumulusci.core.config.sfdx_org_config import SfdxOrgConfig
from cumulusci.core.exceptions import (
    CumulusCIException,
    ScratchOrgException,
    ServiceNotConfigured,
)
from cumulusci.core.sfdx import sfdx

nl = "\n"  # fstrings can't contain backslashes


class ScratchOrgConfig(SfdxOrgConfig):
    """Salesforce DX Scratch org configuration"""

    @property
    def scratch_info(self):
        """Deprecated alias for sfdx_info.

        Will create the scratch org if necessary.
        """
        return self.sfdx_info

    @property
    def days(self):
        return self.config.setdefault("days", 1)

    @property
    def active(self):
        """Check if an org is alive"""
        return self.date_created and not self.expired

    @property
    def expired(self):
        """Check if an org has already expired"""
        return bool(self.expires) and self.expires < datetime.datetime.utcnow()

    @property
    def expires(self):
        if self.date_created:
            return self.date_created + datetime.timedelta(days=int(self.days))

    @property
    def days_alive(self):
        if self.date_created and not self.expired:
            delta = datetime.datetime.utcnow() - self.date_created
            return delta.days + 1

    def create_org(self):
        """Uses sfdx force:org:create to create the org"""
        if not self.config_file:
            raise ScratchOrgException(
                f"Scratch org config {self.name} is missing a config_file"
            )
        if not self.scratch_org_type:
            self.config["scratch_org_type"] = "workspace"

        args = self._build_org_create_args()
        extra_args = os.environ.get("SFDX_ORG_CREATE_ARGS", "")
        p = sfdx(
            f"force:org:create {extra_args}",
            args=args,
            username=None,
            log_note="Creating scratch org",
        )

        stderr = [line.strip() for line in p.stderr_text]
        stdout = [line.strip() for line in p.stdout_text]

        if p.returncode:
            message = f"{FAILED_TO_CREATE_SCRATCH_ORG}: \n{nl.join(stdout)}\n{nl.join(stderr)}"
            raise ScratchOrgException(message)

        re_obj = re.compile("Successfully created scratch org: (.+), username: (.+)")
        username = None
        for line in stdout:
            match = re_obj.search(line)
            if match:
                self.config["org_id"] = match.group(1)
                self.config["username"] = username = match.group(2)
            self.logger.info(line)
        for line in stderr:
            self.logger.error(line)

        if username is None:
            raise ScratchOrgException(
                "SFDX claimed to be successful but there was no username "
                "in the output...maybe there was a gack?"
            )

        self.config["date_created"] = datetime.datetime.utcnow()

        if self.config.get("set_password"):
            self.generate_password()

        # Flag that this org has been created
        self.config["created"] = True

    def _build_org_create_args(self):
        args = ["-f", self.config_file, "-w", "120"]
        devhub = self._choose_devhub()
        if devhub:
            args += ["--targetdevhubusername", devhub]
        if not self.namespaced:
            args += ["-n"]
        if self.noancestors:
            args += ["--noancestors"]
        if self.days:
            args += ["--durationdays", str(self.days)]
        if self.sfdx_alias:
            args += ["-a", self.sfdx_alias]
        with open(self.config_file, "r") as org_def:
            org_def_data = json.load(org_def)
            org_def_has_email = "adminEmail" in org_def_data
        if self.email_address and not org_def_has_email:
            args += [f"adminEmail={self.email_address}"]
        if self.default:
            args += ["-s"]
        instance = self.instance or os.environ.get("SFDX_SIGNUP_INSTANCE")
        if instance:
            args += [f"instance={instance}"]
        return args

    def _choose_devhub(self):
        """Determine which devhub to specify when calling sfdx, if any."""
        # If a devhub was specified via `cci org scratch`, use it.
        # (This will return None if "devhub" isn't set in the org config,
        # in which case sfdx will use its defaultdevhubusername.)
        devhub = self.devhub
        if not devhub and self.keychain is not None:
            # Otherwise see if one is configured via the "devhub" service
            try:
                devhub_service = self.keychain.get_service("devhub")
            except (ServiceNotConfigured, CumulusCIException):
                pass
            else:
                devhub = devhub_service.username
        return devhub

    def generate_password(self):
        """Generates an org password with the sfdx utility."""

        if self.password_failed:
            self.logger.warning("Skipping resetting password since last attempt failed")
            return

        # Set a random password so it's available via cci org info

        p = sfdx(
            "force:user:password:generate",
            self.username,
            log_note="Generating scratch org user password",
        )

        stderr = p.stderr_text.readlines()
        stdout = p.stdout_text.readlines()

        if p.returncode:
            self.config["password_failed"] = True
            # Don't throw an exception because of failure creating the
            # password, just notify in a log message
            self.logger.warning(
                f"Failed to set password: \n{nl.join(stdout)}\n{nl.join(stderr)}"
            )

    def format_org_days(self):
        if self.days_alive:
            org_days = f"{self.days_alive}/{self.days}"
        else:
            org_days = str(self.days)
        return org_days

    def can_delete(self):
        return bool(self.date_created)

    def delete_org(self):
        """Uses sfdx force:org:delete to delete the org"""
        if not self.created:
            self.logger.info(
                "Skipping org deletion: the scratch org has not been created"
            )
            return

        p = sfdx("force:org:delete -p", self.username, "Deleting scratch org")

        output = []
        for line in list(p.stdout_text) + list(p.stderr_text):
            output.append(line)
            if "error" in line.lower():
                self.logger.error(line)
            else:
                self.logger.info(line)

        if p.returncode:
            message = "Failed to delete scratch org"
            raise ScratchOrgException(message)

        # Flag that this org has been deleted
        self.config["created"] = False
        self.config["username"] = None
        self.config["date_created"] = None
        self.config["instance_url"] = None
        self.save()



from cumulusci.tasks.salesforce import BaseSalesforceApiTask


class GetAvailableLicenses(BaseSalesforceApiTask):
    def _run_task(self):
        self.return_values = [
            result["LicenseDefinitionKey"]
            for result in self.sf.query("SELECT LicenseDefinitionKey FROM UserLicense")[
                "records"
            ]
        ]
        licenses = "\n".join(self.return_values)
        self.logger.info(f"Found licenses:\n{licenses}")


class GetAvailablePermissionSetLicenses(BaseSalesforceApiTask):
    def _run_task(self):
        self.return_values = [
            result["PermissionSetLicenseKey"]
            for result in self.sf.query(
                "SELECT PermissionSetLicenseKey FROM PermissionSetLicense"
            )["records"]
        ]
        licenses = "\n".join(self.return_values)
        self.logger.info(f"Found permission set licenses:\n{licenses}")


class GetAvailablePermissionSets(BaseSalesforceApiTask):
    def _run_task(self):
        self.return_values = [
            result["Name"]
            for result in self.sf.query_all("SELECT Name FROM PermissionSet")["records"]
        ]
        permsets = "\n".join(self.return_values)
        self.logger.info(f"Found Permission Sets:\n{permsets}")



lex_locators = {
    "actions": "//runtime_platform_actions-actions-ribbon//ul|//ul[contains(concat(' "
    "',normalize-space(@class),' '),' oneActionsRibbon ')]",
    "app_launcher": {
        "app_link": "//one-app-launcher-modal//one-app-launcher-app-tile//a[.='{}']",
        "button": "//div[contains(@class,'appLauncher')]//button[//div[contains(@class,'slds-icon-waffle')]]",
        "current_app": "//div[contains(@class,'navLeft')]//span[contains(@class,'appName')]/span[text()='{}']",
        "menu": "//div[contains(@class, 'appLauncherMenu')]",
        "tab_link": "//one-app-launcher-modal//one-app-launcher-tab-item//a[.='{}']",
        "view_all": "//div[contains(@class, "
        "'appLauncherMenu')]//button[text()='View All']",
    },
    "body": "//div[contains(@class, 'slds-template__container')]/*",
    "desktop_rendered": "css: "
    "div.desktop.container.oneOne.oneAppLayoutHost[data-aura-rendered-by]",
    "loading_box": "css: div.auraLoadingBox.oneLoadingBox",
    "modal": {
        "button": "//div[contains(@class,'uiModal')]//button[.='{}']",
        "close": "css: button.slds-modal__close",
        "error_messages": "css: div.pageLevelErrors ul.errorsList li",
        "fade_in": "css: div.slds-fade-in-open",
        "field_alert": "//div[contains(@class, 'forceFormPageError')]",
        "has_error": "css: div.forceFormPageError",
        "is_open": "css: div.uiModal div.panel.slds-modal",
        "review_alert": "//a[@force-recordediterror_recordediterror and "
        'text()="{}"]',
    },
    "object": {
        "button": "//div[contains(@class, "
        "'slds-page-header')]//*[self::a[@title='{title}'] or "
        "self::button[@name='{title}']]",
        "field": "//lightning-input[.//label[text()='{}']]//input",
        "field_label": "//label[@for!='' and "
        "text()='{}']|//label[@for!=''][./span[text()='{}']]",
        "field_lookup_link": "//*[@role='option'][.//*[@title='{}']]",
        "field_lookup_value": "//div[contains(@class, "
        "'uiInput')][.//label[contains(@class, "
        "'uiLabel')][.//span[text()='{}']]]//span[contains(@class,'pillText')]",
        "record_type_option": "//div[contains(@class, "
        "'changeRecordTypeOptionRightColumn')]//span[text()='{}']",
    },
    "popup": {
        "link": "//div[contains(@class, 'uiPopupTarget')][contains(@class, "
        "'visible')]//a[@title='{}']"
    },
    "record": {
        "header": {
            "field": "//li[contains(@class, "
            "'slds-page-header__detail-block')][.//span[contains(@class, "
            "'slds-form-element__label')][@title='{}']]",
            "field_value": "//records-lwc-highlights-panel//force-highlights-details-item[.//*[contains(@class, "
            "'slds-text-title') and "
            "text()='{}']]//p[contains(@class, "
            "'fieldComponent')]//*[text()]",
            "field_value_checked": "//li[contains(@class, "
            "'slds-page-header__detail-block')][.//span[contains(@class, "
            "'slds-form-element__label')][@title='{}']]//span[contains(@class, "
            "'uiOutputCheckbox')]//img[@alt='True']",
            "field_value_link": "//records-lwc-highlights-panel//force-highlights-details-item[.//*[.='{}']]//a",
            "field_value_unchecked": "//li[contains(@class, "
            "'slds-page-header__detail-block')][.//span[contains(@class, "
            "'slds-form-element__label')][@title='{}']]//span[contains(@class, "
            "'uiOutputCheckbox')]//img[@alt='False']",
        },
        "related": {
            "button": "//*[@data-component-id='force_relatedListContainer']//article[contains(@class, "
            "'slds-card "
            "slds-card_boundary')][.//span[@title='{}']]//*[text()='{}']",
            "card": "//*[@data-component-id='force_relatedListContainer']//article[.//span[@title='{}']]",
            "count": "//*[@data-component-id='force_relatedListContainer']//article//span[@title='{}']/following-sibling::span",
            "link": "//*[@data-component-id='force_relatedListContainer']//article[.//span[@title='{}']]//*[text()='{}']",
            "popup_trigger": "//*[@data-component-id='force_relatedListContainer']//article[.//span[@title='{}']]//span[text()='Show "
            "Actions']",
        },
    },
    "spinner": "css: div.slds-spinner",
    "tabs": {"tab": ""},
}



import json
import tempfile
import unittest
from pathlib import Path
from unittest import mock

import pytest
import responses

from cumulusci.core.config import ServiceConfig, TaskConfig
from cumulusci.core.exceptions import (
    GithubApiNotFoundError,
    GithubException,
    TaskOptionsError,
)
from cumulusci.tasks.github.publish import PublishSubtree
from cumulusci.tasks.github.tests.util_github_api import GithubApiTestMixin
from cumulusci.tests.util import create_project_config


class TestPublishSubtree(unittest.TestCase, GithubApiTestMixin):
    def setUp(self):
        self.repo_owner = "TestOwner"
        self.repo_name = "TestRepo"
        self.repo_api_url = (
            f"https://api.github.com/repos/{self.repo_owner}/{self.repo_name}"
        )

        self.public_owner = "TestOwner"
        self.public_name = "PublicRepo"
        self.public_repo_url = (
            f"https://api.github.com/repos/{self.public_owner}/{self.public_name}"
        )

        self.project_config = create_project_config(self.repo_name, self.repo_owner)
        self.project_config.keychain.set_service(
            "github",
            "test_alias",
            ServiceConfig(
                {
                    "username": "TestUser",
                    "token": "TestPass",
                    "email": "testuser@testdomain.com",
                }
            ),
        )

    def test_run_task__invalid_version_option_value(self):
        task_config = TaskConfig(
            {
                "options": {
                    "branch": "main",
                    "version": "invalue_value",
                    "create_release": True,
                    "repo_url": self.public_repo_url,
                    "includes": ["tasks/foo.py", "unpackaged/pre/foo/package.xml"],
                }
            }
        )

        with pytest.raises(TaskOptionsError):
            PublishSubtree(self.project_config, task_config)

    @mock.patch("cumulusci.tasks.github.publish.download_extract_github_from_repo")
    @mock.patch("cumulusci.tasks.github.publish.CommitDir")
    def test_run_task_tag_name__explicit_tag_name(self, commit_dir, extract_github):
        with responses.RequestsMock() as rsps:
            rsps.add(
                method=responses.GET,
                url=self.repo_api_url,
                json=self._get_expected_repo(
                    owner=self.repo_owner, name=self.repo_name
                ),
            )
            rsps.add(
                method=responses.GET,
                url=self.public_repo_url,
                json=self._get_expected_repo(
                    owner=self.public_owner, name=self.public_name
                ),
            )
            rsps.add(
                method=responses.GET,
                url=self.repo_api_url + "/git/ref/tags/release/1.0",
                status=200,
                json=self._get_expected_tag_ref("release/1.0", "SHA"),
            )
            rsps.add(
                method=responses.GET,
                url=self.repo_api_url + "/git/tags/SHA",
                json=self._get_expected_tag("release/1.0", "SHA"),
                status=200,
            )
            rsps.add(
                responses.GET,
                self.repo_api_url + "/releases/tags/release/1.0",
                json=self._get_expected_release("release/1.0"),
            )
            rsps.add(
                responses.GET,
                self.public_repo_url + "/git/ref/tags/release/1.0",
                status=404,
            )
            rsps.add(
                responses.POST,
                self.public_repo_url + "/releases",
                json=self._get_expected_release("release"),
            )
            task_config = TaskConfig(
                {
                    "options": {
                        "branch": "main",
                        "tag_name": "release/1.0",
                        "create_release": True,
                        "repo_url": self.public_repo_url,
                        "includes": ["tasks/foo.py", "unpackaged/pre/foo/package.xml"],
                    }
                }
            )
            extract_github.return_value.namelist.return_value = [
                "tasks/foo.py",
                "unpackaged/pre/foo/package.xml",
                "force-app",
            ]

            task = PublishSubtree(self.project_config, task_config)
            task()

            expected_release_body = json.dumps(
                {
                    "tag_name": "release/1.0",
                    "name": "1.0",
                    "body": "",
                    "draft": False,
                    "prerelease": False,
                }
            )
            create_release_call = rsps.calls[7]
            assert create_release_call.request.url == self.public_repo_url + "/releases"
            assert create_release_call.request.method == responses.POST
            assert create_release_call.request.body == expected_release_body

    @mock.patch("cumulusci.tasks.github.publish.download_extract_github_from_repo")
    @mock.patch("cumulusci.tasks.github.publish.CommitDir")
    def test_run_task_tag_name__latest(self, commit_dir, extract_github):
        with responses.RequestsMock() as rsps:
            rsps.add(
                method=responses.GET,
                url=self.repo_api_url,
                json=self._get_expected_repo(
                    owner=self.repo_owner, name=self.repo_name
                ),
            )
            rsps.add(
                method=responses.GET,
                url=self.public_repo_url,
                json=self._get_expected_repo(
                    owner=self.public_owner, name=self.public_name
                ),
            )
            rsps.add(
                method=responses.GET,
                url=self.repo_api_url + "/releases/latest",
                status=200,
                json=self._get_expected_release("release/1.0"),
            )
            rsps.add(
                method=responses.GET,
                url=self.repo_api_url + "/git/ref/tags/release/1.0",
                status=200,
                json=self._get_expected_tag_ref("release/1.0", "SHA"),
            )
            rsps.add(
                method=responses.GET,
                url=self.repo_api_url + "/git/tags/SHA",
                json=self._get_expected_tag("release/1.0", "SHA"),
                status=200,
            )
            rsps.add(
                responses.GET,
                self.repo_api_url + "/releases/tags/release/1.0",
                json=self._get_expected_release("release/1.0"),
            )
            rsps.add(
                responses.GET,
                self.public_repo_url + "/git/ref/tags/release/1.0",
                status=404,
            )
            rsps.add(
                responses.POST,
                self.public_repo_url + "/releases",
                json=self._get_expected_release("release"),
            )
            task_config = TaskConfig(
                {
                    "options": {
                        "branch": "main",
                        "tag_name": "latest",
                        "create_release": True,
                        "repo_url": self.public_repo_url,
                        "includes": ["tasks/foo.py", "unpackaged/pre/foo/package.xml"],
                    }
                }
            )
            extract_github.return_value.namelist.return_value = [
                "tasks/foo.py",
                "unpackaged/pre/foo/package.xml",
                "force-app",
            ]

            task = PublishSubtree(self.project_config, task_config)
            task()

            expected_release_body = json.dumps(
                {
                    "tag_name": "release/1.0",
                    "name": "1.0",
                    "body": "",
                    "draft": False,
                    "prerelease": False,
                }
            )
            create_release_call = rsps.calls[9]
            assert create_release_call.request.url == self.public_repo_url + "/releases"
            assert create_release_call.request.method == responses.POST
            assert create_release_call.request.body == expected_release_body

    @mock.patch("cumulusci.tasks.github.publish.download_extract_github_from_repo")
    @mock.patch("cumulusci.tasks.github.publish.CommitDir")
    def test_run_task_version(self, commit_dir, extract_github):
        with responses.RequestsMock() as rsps:
            rsps.add(
                method=responses.GET,
                url=self.repo_api_url,
                json=self._get_expected_repo(
                    owner=self.repo_owner, name=self.repo_name
                ),
            )
            rsps.add(
                method=responses.GET,
                url=self.public_repo_url,
                json=self._get_expected_repo(
                    owner=self.public_owner, name=self.public_name
                ),
            )
            rsps.add(
                method=responses.GET,
                url=self.repo_api_url + "/releases",
                status=200,
                json=self._get_expected_releases(
                    "TestOwner", "TestRepo", ["beta/1.0-Beta_1"]
                ),
            )
            rsps.add(
                method=responses.GET,
                url=self.repo_api_url + "/git/ref/tags/beta/1.0-Beta_1",
                status=200,
                json=self._get_expected_tag_ref("release/1.0", "SHA"),
            )
            rsps.add(
                method=responses.GET,
                url=self.repo_api_url + "/git/tags/SHA",
                json=self._get_expected_tag("release/1.0", "SHA"),
                status=200,
            )
            rsps.add(
                responses.GET,
                self.repo_api_url + "/releases/tags/beta/1.0-Beta_1",
                json=self._get_expected_release("release/1.0"),
            )
            rsps.add(
                responses.GET,
                self.public_repo_url + "/git/ref/tags/beta/1.0-Beta_1",
                status=404,
            )
            rsps.add(
                responses.POST,
                self.public_repo_url + "/releases",
                json=self._get_expected_release("beta/1.0-Beta_1"),
            )
            task_config = TaskConfig(
                {
                    "options": {
                        "branch": "main",
                        "version": "latest_beta",
                        "create_release": True,
                        "repo_url": self.public_repo_url,
                        "includes": ["tasks/foo.py", "unpackaged/pre/foo/package.xml"],
                    }
                }
            )
            extract_github.return_value.namelist.return_value = [
                "tasks/foo.py",
                "unpackaged/pre/foo/package.xml",
                "force-app",
            ]

            task = PublishSubtree(self.project_config, task_config)
            task()

            expected_release_body = json.dumps(
                {
                    "tag_name": "beta/1.0-Beta_1",
                    "name": "1.0 (Beta 1)",
                    "body": "",
                    "draft": False,
                    "prerelease": False,
                }
            )
            create_release_call = rsps.calls[9]
            assert create_release_call.request.url == self.public_repo_url + "/releases"
            assert create_release_call.request.method == responses.POST
            assert create_release_call.request.body == expected_release_body

    @responses.activate
    @mock.patch("cumulusci.tasks.github.publish.download_extract_github_from_repo")
    @mock.patch("cumulusci.tasks.github.publish.CommitDir")
    def test_ref_not_found(self, commit_dir, extract_github):
        responses.add(
            method=responses.GET,
            url=self.repo_api_url,
            json=self._get_expected_repo(owner=self.repo_owner, name=self.repo_name),
        )
        responses.add(
            method=responses.GET,
            url=self.public_repo_url,
            json=self._get_expected_repo(
                owner=self.public_owner, name=self.public_name
            ),
        )
        responses.add(
            responses.GET,
            self.repo_api_url + "/releases/latest",
            json=self._get_expected_release("release/1.0"),
        )
        responses.add(
            method=responses.GET,
            url=self.repo_api_url + "/git/ref/tags/release/1.0",
            status=404,
        )
        task_config = TaskConfig(
            {
                "options": {
                    "branch": "main",
                    "tag_name": "release/1.0",
                    "create_release": True,
                    "repo_url": self.public_repo_url,
                    "includes": ["tasks/foo.py", "unpackaged/pre/foo/package.xml"],
                }
            }
        )
        extract_github.return_value.namelist.return_value = [
            "tasks/foo.py",
            "unpackaged/pre/foo/package.xml",
            "force-app",
        ]
        task = PublishSubtree(self.project_config, task_config)
        with pytest.raises(GithubApiNotFoundError) as exc:
            task()
        assert "Could not find reference for 'tags/release/1.0' on GitHub" == str(
            exc.value
        )

    @responses.activate
    @mock.patch("cumulusci.tasks.github.publish.download_extract_github_from_repo")
    @mock.patch("cumulusci.tasks.github.publish.CommitDir")
    def test_tag_not_found(self, commit_dir, extract_github):
        responses.add(
            method=responses.GET,
            url=self.repo_api_url,
            json=self._get_expected_repo(owner=self.repo_owner, name=self.repo_name),
        )
        responses.add(
            responses.GET,
            self.repo_api_url + "/releases/latest",
            json=self._get_expected_release("release/1.0"),
        )
        responses.add(
            method=responses.GET,
            url=self.public_repo_url,
            json=self._get_expected_repo(
                owner=self.public_owner, name=self.public_name
            ),
        )
        responses.add(
            method=responses.GET,
            url=self.repo_api_url + "/git/ref/tags/release/1.0",
            status=201,
            json=self._get_expected_tag_ref("release/1.0", "REF_SHA"),
        )
        responses.add(
            responses.GET,
            self.public_repo_url + "/releases/tags/release/1.0",
            json=self._get_expected_release("release/1.0"),
        )
        responses.add(
            method=responses.GET,
            url=self.repo_api_url + "/git/tags/REF_SHA",
            status=404,
        )
        task_config = TaskConfig(
            {
                "options": {
                    "branch": "main",
                    "tag_name": "release/1.0",
                    "create_release": True,
                    "repo_url": self.public_repo_url,
                    "includes": ["tasks/foo.py", "unpackaged/pre/foo/package.xml"],
                }
            }
        )
        extract_github.return_value.namelist.return_value = [
            "tasks/foo.py",
            "unpackaged/pre/foo/package.xml",
            "force-app",
        ]
        task = PublishSubtree(self.project_config, task_config)
        with pytest.raises(GithubApiNotFoundError) as exc:
            task()
        assert "Could not find tag 'release/1.0' with SHA REF_SHA" in str(exc.value)

    @responses.activate
    @mock.patch("cumulusci.tasks.github.publish.download_extract_github_from_repo")
    @mock.patch("cumulusci.tasks.github.publish.CommitDir")
    def test_release_not_found(self, commit_dir, extract_github):
        responses.add(
            method=responses.GET,
            url=self.repo_api_url,
            json=self._get_expected_repo(owner=self.repo_owner, name=self.repo_name),
        )
        responses.add(
            responses.GET,
            self.repo_api_url + "/releases/latest",
            json=self._get_expected_release("release/1.0"),
        )
        responses.add(
            method=responses.GET,
            url=self.public_repo_url,
            json=self._get_expected_repo(
                owner=self.public_owner, name=self.public_name
            ),
        )
        responses.add(
            method=responses.GET,
            url=self.repo_api_url + "/git/ref/tags/release/1.0",
            status=200,
            json=self._get_expected_tag_ref("release/1.0", "SHA"),
        )
        responses.add(
            method=responses.GET,
            url=self.repo_api_url + "/git/tags/SHA",
            json=self._get_expected_tag("release/1.0", "SHA"),
            status=200,
        )
        responses.add(
            responses.GET, self.repo_api_url + "/releases/tags/release/1.0", status=404
        )
        task_config = TaskConfig(
            {
                "options": {
                    "branch": "master",
                    "tag_name": "release/1.0",
                    "create_release": True,
                    "repo_url": self.public_repo_url,
                    "includes": ["tasks/foo.py", "unpackaged/pre/foo/package.xml"],
                }
            }
        )
        extract_github.return_value.namelist.return_value = [
            "tasks/foo.py",
            "unpackaged/pre/foo/package.xml",
            "force-app",
        ]
        task = PublishSubtree(self.project_config, task_config)
        with pytest.raises(GithubException) as exc:
            task()
        assert "Release for release/1.0 not found" == str(exc.value)

    @responses.activate
    @mock.patch("cumulusci.tasks.github.publish.download_extract_github_from_repo")
    @mock.patch("cumulusci.tasks.github.publish.CommitDir")
    def test_target_release_exists(self, commit_dir, extract_github):
        responses.add(
            method=responses.GET,
            url=self.repo_api_url,
            json=self._get_expected_repo(owner=self.repo_owner, name=self.repo_name),
        )
        responses.add(
            responses.GET,
            self.repo_api_url + "/releases/latest",
            json=self._get_expected_release("release/1.0"),
        )
        responses.add(
            method=responses.GET,
            url=self.public_repo_url,
            json=self._get_expected_repo(
                owner=self.public_owner, name=self.public_name
            ),
        )
        responses.add(
            method=responses.GET,
            url=self.public_repo_url + "/git/ref/tags/release/1.0",
            status=201,
            json=self._get_expected_tag_ref("release/1.0", "SHA"),
        )
        responses.add(
            method=responses.GET,
            url=self.repo_api_url + "/git/ref/tags/release/1.0",
            status=201,
            json=self._get_expected_tag_ref("release/1.0", "SHA"),
        )
        responses.add(
            method=responses.GET,
            url=self.repo_api_url + "/git/tags/SHA",
            json=self._get_expected_tag("release/1.0", "SHA"),
            status=201,
        )
        responses.add(
            responses.GET,
            self.repo_api_url + "/releases/tags/release/1.0",
            json=self._get_expected_release("release/1.0"),
        )
        responses.add(
            responses.GET,
            self.public_repo_url + "/releases/tags/release/1.0",
            json=self._get_expected_release("release/1.0"),
        )
        task_config = TaskConfig(
            {
                "options": {
                    "branch": "master",
                    "tag_name": "release/1.0",
                    "create_release": True,
                    "repo_url": self.public_repo_url,
                    "includes": ["tasks/foo.py", "unpackaged/pre/foo/package.xml"],
                }
            }
        )
        extract_github.return_value.namelist.return_value = [
            "tasks/foo.py",
            "unpackaged/pre/foo/package.xml",
            "force-app",
        ]
        task = PublishSubtree(self.project_config, task_config)
        with pytest.raises(GithubException) as exc:
            task()
        assert "Ref for tag release/1.0 already exists in target repo" == str(exc.value)

    def test_ref_nor_tag_name_error(self):
        task_config = TaskConfig(
            {
                "options": {
                    "branch": "master",
                    "repo_url": self.public_repo_url,
                    "includes": ["tasks/foo.py", "unpackaged/pre/foo/package.xml"],
                }
            }
        )
        with pytest.raises(TaskOptionsError) as exc:
            PublishSubtree(self.project_config, task_config)
        assert "Either `ref` or `tag_name` option is required." == str(exc.value)

    def test_renames_not_list_error(self):
        task_config = TaskConfig(
            {
                "options": {
                    "branch": "master",
                    "repo_url": self.public_repo_url,
                    "includes": ["tasks/foo.py", "unpackaged/pre/foo/package.xml"],
                    "renames": ["tasks/foo.py", "unpackaged/pre/foo/package.xml"],
                }
            }
        )
        with pytest.raises(TaskOptionsError) as exc:
            PublishSubtree(self.project_config, task_config)
        assert (
            "Renamed paths must be a list of dicts with `local:` and `target:` keys."
            == str(exc.value)
        )

    def test_renames_wrong_keys_error(self):
        task_config = TaskConfig(
            {
                "options": {
                    "branch": "master",
                    "repo_url": self.public_repo_url,
                    "includes": ["tasks/foo.py", "unpackaged/pre/foo/package.xml"],
                    "renames": [
                        {
                            "wrong_key": "tasks/foo.py",
                            "target": "unpackaged/pre/foo/package.xml",
                        }
                    ],
                }
            }
        )
        with pytest.raises(TaskOptionsError) as exc:
            PublishSubtree(self.project_config, task_config)
        assert (
            "Renamed paths must be a list of dicts with `local:` and `target:` keys."
            == str(exc.value)
        )

    def test_renames_bad_value_error(self):
        task_config = TaskConfig(
            {
                "options": {
                    "branch": "master",
                    "repo_url": self.public_repo_url,
                    "create_release": True,
                    "includes": ["tasks/foo.py", "unpackaged/pre/foo/package.xml"],
                    "renames": [
                        {
                            "local": "public/target_readme.md",
                            "target": "",
                        }
                    ],
                }
            }
        )
        with pytest.raises(TaskOptionsError) as exc:
            PublishSubtree(self.project_config, task_config)
        assert (
            "Renamed paths must be a list of dicts with `local:` and `target:` keys."
            == str(exc.value)
        )

    def test_path_renames_option(self):
        test_includes = [
            "scripts/anon.cls",
            "scripts/more_anon.cls",
            "tasks/unmoved.py",
            "tasks/move_me.py",
            "public/public_readme.md",
        ]
        test_renames = [
            {"local": "scripts", "target": "apex"},
            {"local": "tasks/move_me.py", "target": "sample_tasks/was_moved.py"},
            {"local": "public/public_readme.md", "target": "README.md"},
            {"local": "missing_path", "target": "should_not_exist"},
        ]
        task_config = TaskConfig(
            {
                "options": {
                    "branch": "master",
                    "ref": "some-branch-name",
                    "repo_url": self.public_repo_url,
                    "include": test_includes,
                    "renames": test_renames,
                }
            }
        )
        task = PublishSubtree(self.project_config, task_config)
        with tempfile.TemporaryDirectory() as target:
            included_paths = [Path(target, local_name) for local_name in test_includes]
            [
                include.parent.mkdir(parents=True, exist_ok=True)
                for include in included_paths
            ]

            assert all([include.write_text("12345") == 5 for include in included_paths])
            assert all([include.exists() for include in included_paths])

            # Renames without a local shouldn't exist
            missing_rename = test_renames.pop()
            missing_local = Path(target, missing_rename["local"])
            assert missing_local not in included_paths

            task._rename_files(target)

            assert not Path(target, missing_rename["target"]).exists()
            unmoved_path = Path(target, "tasks/unmoved.py")
            assert unmoved_path.exists()

            # All of the paths with renames should be gone
            included_paths.remove(unmoved_path)
            assert all([not include.exists() for include in included_paths])

            # All of the renamed paths should exist
            new_paths = [Path(target, rename["target"]) for rename in test_renames]
            assert all([rename.exists() for rename in new_paths])

            # Directory renames should work
            assert Path(target, "apex/anon.cls").exists()
            assert Path(target, "apex/more_anon.cls").exists()

    @mock.patch("cumulusci.tasks.github.publish.download_extract_github_from_repo")
    @mock.patch("cumulusci.tasks.github.publish.CommitDir.__call__")
    def test_run_task_ref(self, commit_dir, extract_github):
        with responses.RequestsMock() as rsps:
            rsps.add(
                method=responses.GET,
                url=self.repo_api_url,
                json=self._get_expected_repo(
                    owner=self.repo_owner, name=self.repo_name
                ),
            )
            rsps.add(
                method=responses.GET,
                url=self.public_repo_url,
                json=self._get_expected_repo(
                    owner=self.public_owner, name=self.public_name
                ),
            )
            task_config = TaskConfig(
                {
                    "options": {
                        "branch": "master",
                        "ref": "feature/publish",
                        "create_release": False,
                        "repo_url": self.public_repo_url,
                        "includes": ["tasks/foo.py", "unpackaged/pre/foo/package.xml"],
                    }
                }
            )
            extract_github.return_value.namelist.return_value = [
                "tasks/foo.py",
                "unpackaged/pre/foo/package.xml",
                "force-app",
            ]

            task = PublishSubtree(self.project_config, task_config)
            task()

            extract_github.assert_called_once()
            assert "feature/publish" == extract_github.call_args[1]["ref"]
            commit_dir.assert_called_once()
            expected_commit_message = "Published content from ref feature/publish"
            assert commit_dir.call_args[1]["commit_message"] == expected_commit_message

    def test_included_dirs_match(self):
        test_includes = [
            "orgs/",
            "public/public_readme.md",
            "scripts/anon.cls",
            "scripts/public",
            "tasks/move_me.py",
        ]
        test_namelist = [
            "orgs/",
            "orgs/dev.json",
            "orgs/feature.json",
            "orgs/prerelease.json",
            "orgs/release.json",
            "orgs/trial.json",
            "public/public_readme.md",
            "scripts/",
            "scripts/anon.cls",
            "scripts/more_anon.cls",
            "scripts/public/",
            "scripts/public/anon.cls",
            "scripts/public_to_global.cls",
            "tasks/",
            "tasks/move_me.py",
            "tasks/unmoved.py",
        ]
        task_config = TaskConfig(
            {
                "options": {
                    "branch": "master",
                    "ref": "some-branch-name",
                    "repo_url": self.public_repo_url,
                    "include": test_includes,
                }
            }
        )
        task = PublishSubtree(self.project_config, task_config)
        expected_namelist = [
            "orgs/",
            "orgs/dev.json",
            "orgs/feature.json",
            "orgs/prerelease.json",
            "orgs/release.json",
            "orgs/trial.json",
            "public/public_readme.md",
            "scripts/anon.cls",
            "scripts/public/",
            "scripts/public/anon.cls",
            "tasks/move_me.py",
        ]
        actual_namelist = task._filter_namelist(test_includes, test_namelist)
        assert sorted(expected_namelist) == sorted(actual_namelist)



import logging
from unittest import mock

import pytest
import responses

from cumulusci.core.config import OrgConfig
from cumulusci.core.dependencies.utils import TaskContext
from cumulusci.core.exceptions import PackageInstallError
from cumulusci.salesforce_api.exceptions import MetadataApiError
from cumulusci.salesforce_api.package_install import (
    ApexCompileType,
    NameConflictResolution,
    PackageInstallOptions,
    SecurityType,
    UpgradeType,
    install_package_by_namespace_version,
    install_package_by_version_id,
)
from cumulusci.tests.util import create_project_config


@responses.activate
def test_install_package_by_version_id(caplog):
    caplog.set_level(logging.INFO)
    responses.add(
        "POST",
        "https://salesforce/services/data/v52.0/tooling/sobjects/PackageInstallRequest/",
        json={"id": "0Hf"},
    )
    responses.add(
        "GET",
        "https://salesforce/services/data/v52.0/tooling/query/",
        json={"records": [{"Status": "IN_PROGRESS"}]},
    )
    responses.add(
        "GET",
        "https://salesforce/services/data/v52.0/tooling/query/",
        json={"records": [{"Status": "SUCCESS"}]},
    )

    project_config = create_project_config()
    org_config = OrgConfig(
        {"instance_url": "https://salesforce", "access_token": "TOKEN"}, "test"
    )
    install_package_by_version_id(
        project_config, org_config, "04t", PackageInstallOptions()
    )
    assert "Success" in caplog.text


@responses.activate
def test_install_package_by_version_id__error():
    responses.add(
        "POST",
        "https://salesforce/services/data/v52.0/tooling/sobjects/PackageInstallRequest/",
        json={"id": "0Hf"},
    )
    responses.add(
        "GET",
        "https://salesforce/services/data/v52.0/tooling/query/",
        json={
            "records": [
                {
                    "Status": "ERROR",
                    "Errors": {"errors": [{"message": "We have a problem."}]},
                }
            ]
        },
    )

    project_config = create_project_config()
    org_config = OrgConfig(
        {"instance_url": "https://salesforce", "access_token": "TOKEN"}, "test"
    )
    with pytest.raises(PackageInstallError, match="We have a problem."):
        install_package_by_version_id(
            project_config, org_config, "04t", PackageInstallOptions()
        )


@responses.activate
def test_install_package_by_version_id__not_propagated(caplog):
    caplog.set_level(logging.INFO)
    responses.add(
        "POST",
        "https://salesforce/services/data/v52.0/tooling/sobjects/PackageInstallRequest/",
        json={"id": "0Hf"},
    )
    responses.add(
        "GET",
        "https://salesforce/services/data/v52.0/tooling/query/",
        status=400,
        body="invalid cross reference id",
    )
    responses.add(
        "GET",
        "https://salesforce/services/data/v52.0/tooling/query/",
        json={"records": [{"Status": "SUCCESS"}]},
    )

    project_config = create_project_config()
    org_config = OrgConfig(
        {"instance_url": "https://salesforce", "access_token": "TOKEN"}, "test"
    )
    install_package_by_version_id(
        project_config, org_config, "04t", PackageInstallOptions()
    )
    assert "Retrying" in caplog.text
    assert "Success" in caplog.text


@mock.patch("cumulusci.salesforce_api.package_install.ApiDeploy")
@mock.patch("cumulusci.salesforce_api.package_install.InstallPackageZipBuilder")
def test_install_package_by_namespace_version(zip_builder, api_deploy):
    pc = mock.Mock()
    org = mock.Mock()

    install_package_by_namespace_version(
        pc,
        org,
        "foo",
        "1.0",
        PackageInstallOptions(
            activate_remote_site_settings=True,
            password="foobar",
            security_type=SecurityType.PUSH,
        ),
    )

    task = TaskContext(org_config=org, project_config=pc, logger=mock.ANY)
    zip_builder.assert_called_once_with(
        namespace="foo",
        version="1.0",
        activateRSS=True,
        password="foobar",
        securityType="PUSH",
    )
    api_deploy.assert_called_once_with(task, mock.ANY, purge_on_delete=False)
    api_deploy.return_value.assert_called_once()


@mock.patch("cumulusci.salesforce_api.package_install.ApiDeploy")
@mock.patch("cumulusci.salesforce_api.package_install.InstallPackageZipBuilder")
def test_install_package_by_namespace_version__retry(zip_builder, api_deploy):
    pc = mock.Mock()
    org = mock.Mock()
    api_deploy.return_value.side_effect = [
        MetadataApiError("invalid cross reference id", None),
        None,
    ]

    install_package_by_namespace_version(
        pc,
        org,
        "foo",
        "1.0",
        PackageInstallOptions(),
    )

    task = TaskContext(org_config=org, project_config=pc, logger=mock.ANY)
    api_deploy.assert_has_calls(
        [
            mock.call(task, mock.ANY, purge_on_delete=False),
            mock.call(task, mock.ANY, purge_on_delete=False),
        ],
        any_order=True,
    )

    api_deploy.return_value.assert_has_calls([mock.call(), mock.call()])


def test_package_install_options_from_task_options():
    task_options = {
        "activate_remote_site_settings": "False",
        "name_conflict_resolution": "RenameMetadata",
        "password": "foo",
        "security_type": "PUSH",
    }

    assert PackageInstallOptions.from_task_options(
        task_options
    ) == PackageInstallOptions(
        activate_remote_site_settings=False,
        name_conflict_resolution=NameConflictResolution.RENAME,
        password="foo",
        security_type=SecurityType.PUSH,
    )


def test_package_install_options_from_task_options__omitting_optionals():
    task_options = {
        "activate_remote_site_settings": "False",
        "name_conflict_resolution": "RenameMetadata",
        "password": "foo",
        "security_type": "PUSH",
        "apex_compile_type": "package",
        "upgrade_type": "deprecate-only",
    }

    assert PackageInstallOptions.from_task_options(
        task_options
    ) == PackageInstallOptions(
        activate_remote_site_settings=False,
        name_conflict_resolution=NameConflictResolution.RENAME,
        password="foo",
        security_type=SecurityType.PUSH,
        apex_compile_type=ApexCompileType.PACKAGE,
        upgrade_type=UpgradeType.DEPRECATE_ONLY,
    )



import copy

from cumulusci.robotframework import locators_53

lex_locators = copy.deepcopy(locators_53.lex_locators)
lex_locators["record"]["header"]["field_value"] = (
    "//records-lwc-highlights-panel"
    "//records-highlights-details-item[.//*[contains(@class, 'slds-text-title') and text()='{}']]"
    "//p[contains(@class, 'fieldComponent')]//*[text()]"
)
lex_locators["record"]["header"][
    "field_value_link"
] = "//records-lwc-highlights-panel//records-highlights-details-item[.//*[.='{}']]//a"

lex_locators["modal"][
    "review_alert"
] = "//a[@records-recordediterror_recordediterror and text()='{}']"



import json
import uuid
import zipfile
from collections import defaultdict
from pathlib import Path

import requests

from cumulusci.core.exceptions import DeploymentException
from cumulusci.core.utils import process_list_of_pairs_dict_arg
from cumulusci.utils import temporary_dir
from cumulusci.utils.http.requests_utils import safe_json_from_response

from .base import BaseMarketingCloudTask

MCPM_ENDPOINT = "https://spf.{}.marketingcloudapps.com/api"

PAYLOAD_CONFIG_VALUES = {"preserveCategories": True}

PAYLOAD_NAMESPACE_VALUES = {
    "category": "",
    "prepend": "",
    "append": "",
    "timestamp": True,
}

DEPLOY_FINISHED_STATUS = "DONE"
DEPLOY_ERROR_STATUS = "FATAL_ERROR"


class MarketingCloudDeployTask(BaseMarketingCloudTask):

    task_options = {
        "package_zip_file": {
            "description": "Path to the package zipfile that will be deployed.",
            "required": True,
        },
        "custom_inputs": {
            "description": "Specify custom inputs to the deployment task. Takes a mapping from input key to input value (e.g. 'companyName:Acme,companyWebsite:https://www.salesforce.org:8080').",
            "required": False,
        },
        "name": {
            "description": "The name to give to this particular deploy call. Defaults to a universally unique identifier.",
            "required": False,
        },
        "endpoint": {
            "description": "Override the default endpoint for the Marketing Cloud package manager API (optional)",
            "required": False,
        },
    }

    def _init_options(self, kwargs):
        super()._init_options(kwargs)
        custom_inputs = self.options.get("custom_inputs")
        self.custom_inputs = (
            process_list_of_pairs_dict_arg(custom_inputs) if custom_inputs else None
        )

    def _run_task(self):
        pkg_zip_file = Path(self.options["package_zip_file"])
        if not pkg_zip_file.is_file():
            self.logger.error(f"Package zip file not valid: {pkg_zip_file.name}")
            return

        with temporary_dir(chdir=False) as temp_dir:
            with zipfile.ZipFile(pkg_zip_file) as zf:
                zf.extractall(temp_dir)
                payload = self._construct_payload(Path(temp_dir), self.custom_inputs)

        self.headers = {
            "Authorization": f"Bearer {self.mc_config.access_token}",
            "SFMC-TSSD": self.mc_config.tssd,
        }
        custom_endpoint = self.options.get("endpoint")
        self.endpoint = (
            custom_endpoint
            if custom_endpoint
            else MCPM_ENDPOINT.format(self.get_mc_stack_key())
        )

        self.logger.info(f"Deploying package to: {self.endpoint}/deployments")
        response = requests.post(
            f"{self.endpoint}/deployments",
            json=payload,
            headers=self.headers,
        )
        result = safe_json_from_response(response)

        self.job_id = result["id"]
        self.logger.info(f"Started job {self.job_id}")
        self._poll()

    def _poll_action(self):
        """
        Poll something and process the response.
        Set `self.poll_complete = True` to break polling loop.
        """
        response = requests.get(
            f"{self.endpoint}/deployments/{self.job_id}", headers=self.headers
        )
        result = safe_json_from_response(response)
        self.logger.info(f"Waiting [{result['status']}]...")
        if result["status"] == DEPLOY_FINISHED_STATUS:
            self.poll_complete = True
            self._validate_response(result)
        elif result["status"] == DEPLOY_ERROR_STATUS:
            self.poll_complete = True
            self._report_fatal_error(result)

    def _construct_payload(self, dir_path, custom_inputs=None):
        dir_path = Path(dir_path)
        assert dir_path.is_dir(), "package_directory must be a directory"

        payload = defaultdict(lambda: defaultdict(dict))
        payload["namespace"] = PAYLOAD_NAMESPACE_VALUES
        payload["config"] = PAYLOAD_CONFIG_VALUES
        payload["name"] = self.options.get("name", str(uuid.uuid4()))

        with open(dir_path / "references.json", "r") as f:
            payload["references"] = json.load(f)

        with open(dir_path / "input.json", "r") as f:
            payload["input"] = json.load(f)

        entities_dir = Path(f"{dir_path}/entities")
        for item in entities_dir.glob("**/*.json"):
            if item.is_file():
                entity_name = item.parent.name
                entity_id = item.stem
                with open(item, "r") as f:
                    payload["entities"][entity_name][entity_id] = json.load(f)

        if custom_inputs:
            payload = self._add_custom_inputs_to_payload(custom_inputs, payload)

        return payload

    def _add_custom_inputs_to_payload(self, custom_inputs, payload):
        for input_name, value in custom_inputs.items():
            found = False
            for input in payload["input"]:
                if input["key"] == input_name:
                    input["value"] = value
                    found = True
                    break

            if not found:
                raise DeploymentException(
                    f"Custom input of key {input_name} not found in package."
                )

        return payload

    def _validate_response(self, deploy_info: dict):
        """Checks for any errors present in the response to the deploy request.
        Displays errors if present, else informs use that the deployment was successful."""
        has_error = False
        for entity, info in deploy_info["entities"].items():
            if not info:
                continue

            for entity_id, info in info.items():
                if info["status"] not in ("SUCCESS", "REUSED"):
                    has_error = True
                    self.logger.error(
                        f"Failed to deploy {entity}/{entity_id}. Status: {info['status']}. Issues: {info['issues']}"
                    )

        if has_error:
            raise DeploymentException("Marketing Cloud reported deployment failures.")

        self.logger.info("Deployment completed successfully.")

    def _report_fatal_error(self, result: dict):
        self.logger.error(
            f"> {DEPLOY_ERROR_STATUS} received. Dumping response from Marketing Cloud:\n{result}"
        )
        raise DeploymentException(
            f"Marketing Cloud deploy finished with status of: {DEPLOY_ERROR_STATUS}"
        )



import os
import sys

from simple_salesforce import api, bulk

__import__("pkg_resources").declare_namespace("cumulusci")

__location__ = os.path.dirname(os.path.realpath(__file__))

with open(os.path.join(__location__, "version.txt")) as f:
    __version__ = f.read().strip()

if sys.version_info < (3, 8):  # pragma: no cover
    raise Exception("CumulusCI requires Python 3.8+.")

api.OrderedDict = dict
bulk.OrderedDict = dict



import os
import unittest
from unittest import mock

from cumulusci.core.config import (
    BaseProjectConfig,
    OrgConfig,
    TaskConfig,
    UniversalConfig,
)
from cumulusci.tasks.metadata.package import (
    BaseMetadataParser,
    BundleParser,
    BusinessProcessParser,
    CustomLabelsParser,
    CustomObjectParser,
    DocumentParser,
    LWCBundleParser,
    MetadataFilenameParser,
    MetadataFolderParser,
    MetadataParserMissingError,
    MetadataXmlElementParser,
    MissingNameElementError,
    PackageXmlGenerator,
    ParserConfigurationError,
    RecordTypeParser,
    UpdatePackageXml,
    metadata_sort_key,
)
from cumulusci.utils import temporary_dir, touch

__location__ = os.path.dirname(os.path.realpath(__file__))


class TestPackageXmlGenerator(unittest.TestCase):
    def test_metadata_sort_key(self):
        md = ["a__Test__c", "Test__c"]
        md.sort(key=metadata_sort_key)
        self.assertEqual(["Test__c", "a__Test__c"], md)

    def test_package_name_urlencoding(self):
        api_version = "36.0"
        package_name = "Test & Package"

        expected = '<?xml version="1.0" encoding="UTF-8"?>\n'
        expected += '<Package xmlns="http://soap.sforce.com/2006/04/metadata">\n'
        expected += "    <fullName>Test %26 Package</fullName>\n"
        expected += "    <version>{}</version>\n".format(api_version)
        expected += "</Package>"

        with temporary_dir() as path:
            generator = PackageXmlGenerator(path, api_version, package_name)
            package_xml = generator()

        self.assertEqual(package_xml, expected)

    def test_namespaced_report_folder(self):
        api_version = "36.0"
        package_name = "Test Package"
        test_dir = "namespaced_report_folder"

        path = os.path.join(__location__, "package_metadata", test_dir)

        generator = PackageXmlGenerator(path, api_version, package_name)
        with open(os.path.join(path, "package.xml"), "r") as f:
            expected_package_xml = f.read().strip()
        package_xml = generator()

        self.assertEqual(package_xml, expected_package_xml)

    def test_delete_namespaced_report_folder(self):
        api_version = "36.0"
        package_name = "Test Package"
        test_dir = "namespaced_report_folder"

        path = os.path.join(__location__, "package_metadata", test_dir)

        generator = PackageXmlGenerator(path, api_version, package_name, delete=True)
        with open(os.path.join(path, "destructiveChanges.xml"), "r") as f:
            expected_package_xml = f.read().strip()
        package_xml = generator()

        self.assertEqual(package_xml, expected_package_xml)

    def test_parse_types_unknown_md_type(self):
        with temporary_dir() as path:
            os.mkdir(os.path.join(path, "bogus"))
            generator = PackageXmlGenerator(path, "43.0", "Test Package")
            with self.assertRaises(MetadataParserMissingError):
                generator.parse_types()

    def test_render_xml__managed(self):
        with temporary_dir() as path:
            generator = PackageXmlGenerator(
                path,
                "43.0",
                "Test Package",
                managed=True,
                install_class="Install",
                uninstall_class="Uninstall",
            )
            result = generator()
            self.assertEqual(EXPECTED_MANAGED, result)


EXPECTED_MANAGED = """<?xml version="1.0" encoding="UTF-8"?>
<Package xmlns="http://soap.sforce.com/2006/04/metadata">
    <fullName>Test Package</fullName>
    <postInstallClass>Install</postInstallClass>
    <uninstallClass>Uninstall</uninstallClass>
    <version>43.0</version>
</Package>"""


class TestBaseMetadataParser(unittest.TestCase):
    def test_parse_items__skips_files(self):
        with temporary_dir() as path:
            # create files that should be ignored by the parser
            for filename in (
                ".hidden",
                "CODEOWNERS",
                "OWNERS",
                "test.wrongext",
                "test-meta.xml",
                "Account.object",
                "Custom__c.object",
            ):
                touch(filename)

            parser = BaseMetadataParser("TestMDT", path, "object", delete=True)
            parser.parse_item = mock.Mock()
            parser.parse_items()
            parser.parse_item.assert_called_once()

    def test_check_delete_excludes__not_deleting(self):
        parser = BaseMetadataParser("TestMDT", None, "object", delete=False)
        self.assertFalse(parser.check_delete_excludes("asdf"))

    def test_parse_item(self):
        parser = BaseMetadataParser("TestMDT", None, "object", delete=False)
        with self.assertRaises(NotImplementedError):
            parser._parse_item("asdf")

    def test_render_xml__no_members(self):
        parser = BaseMetadataParser("TestMDT", None, "object", delete=False)
        self.assertIsNone(parser.render_xml())


class TestMetadataFilenameParser(unittest.TestCase):
    def test_parse_item(self):
        parser = MetadataFilenameParser("TestMDT", None, "object", delete=False)
        result = parser._parse_item("Test.object")
        self.assertEqual(["Test"], result)

    def test_parse_item_translates_namespace_tokens(self):
        with temporary_dir() as path:
            touch("___NAMESPACE___Foo__c.object")
            parser = MetadataFilenameParser("TestMDT", path, "object", delete=False)
            parser.parse_items()
        self.assertEqual(["%%%NAMESPACE%%%Foo__c"], parser.members)


class TestMetadataFolderParser(unittest.TestCase):
    def test_parse_item(self):
        with temporary_dir() as path:
            item_path = os.path.join(path, "Test")
            os.mkdir(item_path)
            other_path = os.path.join(path, "FolderWithoutMetaXml")
            os.mkdir(other_path)
            with open(os.path.join(path, "Test-meta.xml"), "w"):
                pass
            with open(os.path.join(item_path, ".hidden"), "w"):
                pass
            with open(os.path.join(item_path, "Test.object"), "w"):
                pass
            parser = MetadataFolderParser("TestMDT", path, "object", delete=False)
            self.assertEqual(["Test", "Test/Test"], parser._parse_item("Test"))

    def test_parse_item__non_directory(self):
        with temporary_dir() as path:
            with open(os.path.join(path, "file"), "w"):
                pass
            parser = MetadataFolderParser("TestMDT", path, "object", delete=False)
            self.assertEqual([], parser._parse_item("file"))


class TestBundleParser(unittest.TestCase):
    def test_parse_item(self):
        with temporary_dir() as path:
            item_path = os.path.join(path, "Test")
            os.mkdir(item_path)
            with open(os.path.join(item_path, ".hidden"), "w"):
                pass
            # subitems should be ignored
            with open(os.path.join(item_path, "Test.object"), "w"):
                pass
            parser = BundleParser("TestMDT", path, "object", delete=False)
            self.assertEqual(["Test"], parser._parse_item("Test"))

    def test_parse_item__non_directory(self):
        with temporary_dir() as path:
            with open(os.path.join(path, "file"), "w"):
                pass
            parser = BundleParser("TestMDT", path, "object", delete=False)
            self.assertEqual([], parser._parse_item("file"))


class TestLWCBundleParser(unittest.TestCase):
    def test_parse_item(self):
        with temporary_dir() as path:
            item_path = os.path.join(path, "Test")
            os.mkdir(item_path)
            with open(os.path.join(item_path, ".hidden"), "w"):
                pass
            # subitems should be ignored
            with open(os.path.join(item_path, "Test.object"), "w"):
                pass
            parser = LWCBundleParser(
                "LightningComponentBundle", path, "object", delete=False
            )
            self.assertEqual(["Test"], parser._parse_item("Test"))

    def test_parse_item__no_tests_mocks(self):
        with temporary_dir() as path:
            # Test and mock directories should be ignored.
            item_path = os.path.join(path, "__tests__")
            os.mkdir(item_path)
            parser = LWCBundleParser(
                "LightningComponentBundle", path, None, delete=False
            )
            self.assertEqual([], parser._parse_item("__tests__"))

    def test_parse_item__non_directory(self):
        with temporary_dir() as path:
            with open(os.path.join(path, "file"), "w"):
                pass
            parser = LWCBundleParser(
                "LightningComponentBundle", path, None, delete=False
            )
            self.assertEqual([], parser._parse_item("file"))


class TestMetadataXmlElementParser(unittest.TestCase):
    def test_parser(self):
        with temporary_dir() as path:
            with open(os.path.join(path, "Test.test"), "w") as f:
                f.write(
                    """<?xml version='1.0' encoding='utf-8'?>
<root xmlns="http://soap.sforce.com/2006/04/metadata">
    <test>
        <fullName>Test</fullName>
    </test>
</root>"""
                )
            parser = MetadataXmlElementParser(
                "TestMDT", path, "test", delete=False, item_xpath="./sf:test"
            )
            result = parser()
            self.assertEqual(
                """    <types>
        <members>Test.Test</members>
        <name>TestMDT</name>
    </types>""",
                "\n".join(result),
            )

    def test_parser__missing_item_xpath(self):
        with self.assertRaises(ParserConfigurationError):
            parser = MetadataXmlElementParser("TestMDT", None, "test", False)
            self.assertIsNotNone(parser)

    def test_parser__missing_name(self):
        with temporary_dir() as path:
            with open(os.path.join(path, "Test.test"), "w") as f:
                f.write(
                    """<?xml version='1.0' encoding='utf-8'?>
<root xmlns="http://soap.sforce.com/2006/04/metadata">
    <test />
</root>"""
                )
            parser = MetadataXmlElementParser(
                "TestMDT", path, "test", delete=False, item_xpath="./sf:test"
            )
            with self.assertRaises(MissingNameElementError):
                parser()


class TestCustomLabelsParser(unittest.TestCase):
    def test_parser(self):
        with temporary_dir() as path:
            with open(os.path.join(path, "custom.labels"), "w") as f:
                f.write(
                    """<?xml version='1.0' encoding='utf-8'?>
<root xmlns="http://soap.sforce.com/2006/04/metadata">
    <labels>
        <fullName>TestLabel</fullName>
    </labels>
</root>"""
                )
            parser = CustomLabelsParser(
                "CustomLabels", path, "labels", False, item_xpath="./sf:labels"
            )
            self.assertEqual(["TestLabel"], parser._parse_item("custom.labels"))


class TestCustomObjectParser(unittest.TestCase):
    def test_parse_item(self):
        parser = CustomObjectParser("CustomObject", None, "object", False)
        self.assertEqual(["Test__c"], parser._parse_item("Test__c.object"))

    def test_parse_item__skips_namespaced(self):
        parser = CustomObjectParser("CustomObject", None, "object", False)
        self.assertEqual([], parser._parse_item("ns__Object__c.object"))

    def test_parse_item__skips_standard(self):
        parser = CustomObjectParser("CustomObject", None, "object", False)
        self.assertEqual([], parser._parse_item("Account.object"))


class TestRecordTypeParser(unittest.TestCase):
    def test_check_delete_excludes(self):
        parser = RecordTypeParser(
            "RecordType", None, "object", True, "./sf:recordTypes"
        )
        self.assertTrue(parser.check_delete_excludes("asdf"))


class TestBusinessProcessParser(unittest.TestCase):
    def test_check_delete_excludes(self):
        parser = BusinessProcessParser(
            "BusinessProcess", None, "object", True, "./sf:businessProcesses"
        )
        self.assertTrue(parser.check_delete_excludes("asdf"))


class TestDocumentParser(unittest.TestCase):
    def test_parse_subitem(self):
        parser = DocumentParser("Document", None, None, False)
        self.assertEqual(["folder/doc"], parser._parse_subitem("folder", "doc"))


class TestUpdatePackageXml(unittest.TestCase):
    def test_run_task(self):
        src_path = os.path.join(
            __location__, "package_metadata", "namespaced_report_folder"
        )
        with open(os.path.join(src_path, "package.xml"), "r") as f:
            expected = f.read()
        with temporary_dir() as path:
            output_path = os.path.join(path, "package.xml")
            project_config = BaseProjectConfig(
                UniversalConfig(),
                {
                    "project": {
                        "package": {"name": "Test Package", "api_version": "36.0"}
                    }
                },
            )
            task_config = TaskConfig(
                {"options": {"path": src_path, "output": output_path, "managed": True}}
            )
            org_config = OrgConfig({}, "test")
            task = UpdatePackageXml(project_config, task_config, org_config)
            task()
            with open(output_path, "r") as f:
                result = f.read()
            self.assertEqual(expected, result)



import csv
import os
import os.path
import re
import time
from pathlib import Path

import jinja2
import robot.utils
from robot.libdocpkg.builder import DocumentationBuilder
from robot.libdocpkg.robotbuilder import LibraryDocBuilder
from robot.libraries.BuiltIn import RobotNotRunningError
from robot.utils import Importer

import cumulusci
from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.core.tasks import BaseTask
from cumulusci.core.utils import process_bool_arg, process_glob_list_arg
from cumulusci.robotframework import PageObjects
from cumulusci.utils.fileutils import view_file


class RobotLibDoc(BaseTask):
    task_options = {
        "path": {
            "description": (
                "The path to one or more keyword libraries to be documented. "
                "The path can be single a python file, a .robot file, a python "
                "module (eg: cumulusci.robotframework.Salesforce) or a comma "
                "separated list of any of those. Glob patterns are supported "
                "for filenames (eg: ``robot/SAL/doc/*PageObject.py``). The order "
                "of the files will be preserved in the generated documentation. "
                "The result of pattern expansion will be sorted"
            ),
            "required": True,
        },
        "output": {
            "description": (
                "The output file where the documentation will be written. "
                "Normally an HTML file will be generated. If the filename "
                "ends with '.csv' then a csv file will be generated instead."
            ),
            "required": True,
        },
        "title": {
            "description": "A string to use as the title of the generated output",
            "required": False,
        },
        "preview": {
            "description": (
                "If True, automatically open a window to view the "
                "generated data when the task is successful"
            ),
            "required": False,
        },
    }

    def _validate_options(self):
        super(RobotLibDoc, self)._validate_options()
        self.options["path"] = process_glob_list_arg(self.options["path"])

        # Attempt to collect all files that don't match existing
        # files. Note: "path" could be a library module path (for example,
        # cumulusci.robotframework.CumulusCI) so we only do this check for
        # files that end in known library suffixes (.py, .robot, .resource).
        bad_files = []
        for path in self.options["path"]:
            name, extension = os.path.splitext(path)
            if extension in (".py", ".robot", ".resource") and not os.path.exists(path):
                bad_files.append(path)

        if bad_files:
            if len(bad_files) == 1:
                error_message = "Unable to find the input file '{}'".format(
                    bad_files[0]
                )
            else:
                files = ", ".join(["'{}'".format(filename) for filename in bad_files])
                error_message = "Unable to find the following input files: {}".format(
                    files
                )
            raise TaskOptionsError(error_message)

        self.options["preview"] = process_bool_arg(self.options.get("preview", False))

    def is_pageobject_library(self, path):
        """Return True if the file looks like a page object library"""
        if path.endswith(".py"):
            with open(path, "r") as f:
                data = f.read()
                if re.search(r"@pageobject\(", data):
                    return True
        return False

    def _run_task(self):
        kwfiles = []
        processed_files = []
        for library_name in self.options["path"]:
            kwfile = KeywordFile(library_name)
            try:
                if self.is_pageobject_library(library_name):
                    PageObjects._reset()
                    module = Importer().import_class_or_module_by_path(
                        os.path.abspath(library_name)
                    )
                    kwfile.doc = module.__doc__
                    if hasattr(module, "TITLE"):
                        kwfile.title = module.TITLE

                    for pobj_name, pobj in sorted(PageObjects.registry.items()):
                        pobj = PageObjects.registry[pobj_name]
                        libname = "{}.{}".format(pobj.__module__, pobj.__name__)
                        libdoc = LibraryDocBuilder().build(libname)
                        libdoc.src = os.path.basename(library_name)
                        libdoc.pobj = libname
                        kwfile.add_keywords(libdoc, pobj_name)

                else:
                    libdoc = DocumentationBuilder(library_name).build(library_name)
                    kwfile.add_keywords(libdoc)

                # if we get here, we were able to process the file correctly
                kwfiles.append(kwfile)
                processed_files.append(library_name)

            except RobotNotRunningError as e:
                # oddly, robot's exception has a traceback embedded in the message, so we'll
                # only print out the first line to hide most of the noise
                self.logger.warning(
                    "unexpected error: {}".format(str(e).split("\n")[0])
                )

        try:
            if self.options["output"].endswith(".csv"):
                data = self._convert_to_tuples(kwfiles)
                with open(self.options["output"], "w", newline="") as f:
                    csv_writer = csv.writer(f)
                    csv_writer.writerows(data)

            else:
                html = self._convert_to_html(kwfiles)
                with open(self.options["output"], "w") as f:
                    f.write(html)
            self.logger.info("created {}".format(self.options["output"]))

            if self.options["preview"]:
                view_file(self.options["output"])

        except Exception as e:
            raise TaskOptionsError(
                "Unable to create output file '{}' ({})".format(
                    self.options["output"], str(e)
                )
            )

        return {"files": processed_files}

    def _convert_to_tuples(self, kwfiles):
        """Convert the list of keyword files into a list of tuples

        The first element in the list will be a list of column headings.
        """
        rows = []
        for kwfile in kwfiles:
            rows.extend(kwfile.to_tuples())
        rows = sorted(set(rows))
        rows.insert(0, KeywordFile.get_header())
        return rows

    def _convert_to_html(self, kwfiles):
        """Convert the list of keyword files into html"""

        title = self.options.get("title", "Keyword Documentation")
        date = time.strftime("%A %B %d, %I:%M %p")
        cci_version = cumulusci.__version__

        stylesheet_path = os.path.join(os.path.dirname(__file__), "stylesheet.css")
        with open(stylesheet_path) as f:
            stylesheet = f.read()

        jinjaenv = jinja2.Environment(
            loader=jinja2.FileSystemLoader(os.path.dirname(__file__)), autoescape=False
        )
        jinjaenv.filters["robot_html"] = robot.utils.html_format
        template = jinjaenv.get_template("template.html")
        return template.render(
            libraries=kwfiles,
            title=title,
            cci_version=cci_version,
            stylesheet=stylesheet,
            date=date,
        )


class KeywordFile:
    """Helper class which represents a file and its keywords

    A file may have just a bunch of keywords, or groups of
    keywords organized as page objects. Each group of keywords
    is stored in self.keywords, with the page object metadata
    as a key.

    For normal libraries, the key is an empty tuple.
    """

    def __init__(self, path):
        if os.path.exists(path):
            self.filename = os.path.basename(path)
        else:
            # if it's not a file, it must be a module
            self.filename = path.split(".")[-1]
        self.title = self.filename
        self.path = path
        self.keywords = {}

    @classmethod
    def get_header(cls):
        return ["Name", "Source", "Line#", "po type", "po_object", "Documentation"]

    def add_keywords(self, libdoc, page_object=tuple()):
        self.keywords[page_object] = libdoc

    def to_tuples(self):
        """Convert the dictionary of keyword data to a set of tuples"""
        rows = []
        cwd = Path.cwd()
        base_pageobjects_path = os.path.join(
            "cumulusci", "robotframework", "pageobjects"
        )

        for po, libdoc in self.keywords.items():
            (po_type, po_object) = po if po else ("", "")
            for keyword in libdoc.keywords:
                # we don't want to see the same base pageobject
                # keywords a kajillion times. This should probably
                # be configurable, but I don't need it to be right now.
                if base_pageobjects_path in keyword.source:
                    continue

                path = Path(keyword.source)
                if path.is_absolute():
                    try:
                        path = path.relative_to(cwd)
                    except ValueError:
                        # ok, fine. We'll use the path as-is.
                        pass
                path = str(path)

                # make sure that if you change the list of columns here
                # that you modify the `get_header` property too!
                row = (
                    keyword.name,
                    path,
                    keyword.lineno,
                    po_type,
                    po_object,
                    keyword.doc,
                )
                rows.append(row)

        rows = set(rows)
        return rows



from contextlib import contextmanager
from pathlib import Path
from unittest import mock

import pytest

import cumulusci
from cumulusci.cli.org import org_remove, org_scratch, org_scratch_delete
from cumulusci.cli.runtime import CliRuntime
from cumulusci.core.config import OrgConfig, TaskConfig
from cumulusci.core.exceptions import OrgNotFound
from cumulusci.salesforce_api.utils import get_simple_salesforce_connection
from cumulusci.tests.util import unmock_env


def pytest_addoption(parser, pluginmanager):
    """Pytest magic method: add --org, --replace-vcrs and --run-slow-tests features"""
    parser.addoption("--org", action="store", default=None, help="org to use")
    parser.addoption(
        "--run-slow-tests",
        action="store_true",
        default=False,
        help="Include slow tests.",
    )
    parser.addoption(
        "--replace-vcrs",
        action="store_true",
        default=False,
        help="Replace VCR files.",
    )


def pytest_configure(config):
    """Pytest magic method: add pytest markers/decorators"""
    markers = [
        "slow(): a slow test that should only be executed when requested with --run-slow-tests",
        "large_vcr(): a network-based test that generates VCR cassettes too large for version control. Use --org to generate them locally.",
        "needs_org(): a test that needs an org (or at least access to the network) but should not attempt to store VCR cassettes",
        "org_shape(org_name, init_flow): a test that needs a particular org shape",
    ]

    for marker in markers:
        # register additional markers
        config.addinivalue_line("markers", marker)


def pytest_runtest_setup(item):
    """Pytest magic method"""
    marker_names = set(marker.name for marker in item.iter_markers())
    classify_and_modify_test(item, marker_names)


def sf_pytest_cli_orgname(request):
    """Helper to see what org the user has asked for"""
    return request.config.getoption("--org")


@pytest.fixture(scope="session")
def user_requested_network_access(request):
    """Did the user ask to enable network access by specifying an org?"""
    return bool(sf_pytest_cli_orgname(request))


@pytest.fixture(scope="session")
def user_requested_cassette_replacement(request):
    """Did the user ask to delete VCR cassettes?"""
    return request.config.getoption("--replace-vcrs")


@pytest.fixture(scope="session")
def runtime():
    """Get the CumulusCI runtime for the current working directory."""
    return CliRuntime()


@pytest.fixture(scope="session")
def project_config(runtime):
    """Get the project config for the current working directory."""
    return runtime.project_config


@pytest.fixture(scope="session")
def cli_org_config(request):
    """What org did the user specify on the CLI or pytest options?"""
    cli_org_name = sf_pytest_cli_orgname(request)
    if cli_org_name:
        with unmock_env():  # restore real homedir
            runtime = CliRuntime()
            runtime.keychain._load_orgs()
            org_name, org_config = runtime.get_org(cli_org_name)
            assert org_config.scratch, "You should only run tests against scratch orgs."
            org_config.refresh_oauth_token(runtime.keychain)
            return org_config


@pytest.fixture(scope="function")
def org_config(request, current_org_shape, cli_org_config, fallback_org_config):
    """Get an org config with an active access token.

    If an org was requested by the org_shape feature, it is used.

    Otherwise, you can specify the org name using the --org option when running pytest.

    If there was no org provided by those mechanisms, it will use a dummy org.
    """
    org_config = current_org_shape.org_config or cli_org_config
    if org_config:
        org_config.user_specified = True
    else:
        org_config = fallback_org_config()

    # fallback_org_config can be defined in "conftest" based
    # on the needs of the test suite. For example, for
    # fast running test suites it might return a hardcoded
    # org and for integration test suites it might return
    # a specific default org or throw an exception.
    with mock.patch.object(OrgConfig, "latest_api_version", "48.0"), mock.patch.object(
        OrgConfig, "refresh_oauth_token"
    ):
        yield org_config


@pytest.fixture
def sf(request, project_config, org_config):
    """Get a simple-salesforce client for org_config."""
    sf = get_simple_salesforce_connection(project_config, org_config)
    return sf


@pytest.fixture
def create_task(request, project_config, org_config):
    """Get a task _factory_ which can be used to construct task instances."""
    session_project_config = project_config
    session_org_config = org_config

    def create_task(task_class, options=None, project_config=None, org_config=None):
        project_config = project_config or session_project_config
        org_config = org_config or session_org_config
        options = options or {}

        task_config = TaskConfig({"options": options})

        return task_class(project_config, task_config, org_config)

    return create_task


def vcr_cassette_name_for_item(item):
    """Name of the VCR cassette"""
    test_class = item.cls
    if test_class:
        return "{}.{}".format(test_class.__name__, item.name)
    return item.name


def classify_and_modify_test(item, marker_names):
    if "slow" in marker_names:
        if not item.config.getoption("--run-slow-tests"):
            pytest.skip("slow: test requires --run-slow-tests")

    if "large_vcr" in marker_names:
        library_dir = Path(cumulusci.__file__).parent.parent / "large_cassettes"
        item.add_marker(pytest.mark.vcr(cassette_library_dir=str(library_dir)))

        test_name = vcr_cassette_name_for_item(item)
        test_path = Path(library_dir) / (test_name + ".yaml")
        if not (test_path.exists() or item.config.getoption("--org")):
            pytest.skip("large_vcr: test requires --org to generate large cassette")

    elif "needs_org" in marker_names and not item.config.getoption("--org"):
        pytest.skip("needs_org: test requires --org")


class CurrentOrg:
    __slots__ = ("org_config",)

    def __init__(self):
        self.org_config = None


@pytest.fixture(scope="session")
def org_shapes():
    org_shapes = {}
    try:
        yield org_shapes
    finally:
        cleanup_org_shapes(org_shapes)


def cleanup_org_shapes(org_shapes: dict):
    runtime = CliRuntime(load_keychain=True)
    errors = []
    for org_name in org_shapes:
        cleanup_org(runtime, org_name, errors)

    if errors:
        raise AssertionError(str(errors))


def cleanup_org(runtime, org_name, errors):
    try:
        org_scratch_delete.callback.__wrapped__(
            runtime,
            org_name,
        )
    except Exception as e:
        print(f"EXCEPTION deleting org {e}")
        errors.append(e)
    try:
        org_remove.callback.__wrapped__(runtime, org_name, False)
    except Exception:
        pass


@pytest.fixture(scope="session")
def current_org_shape(request):
    """Internal: shared state for fixtures"""
    org = CurrentOrg()
    return org


@pytest.fixture(scope="function", autouse=True)
def _change_org_shape(request, current_org_shape, org_shapes):
    """Select an org_shape for a test
    e.g.:

        @pytest.mark.org_shape("qa", "qa_org")
        def test_foo(create_task);
            t = create_task(Foo)
            t()

    Switch the current org to an org created with
    org template "qa" after running flow "qa_org".

    Clean up any changes you make, because this org may be reused by
    other tests.
    """
    marker = request.node.get_closest_marker("org_shape")
    if marker:
        config_name, flow_name = marker.args
        with change_org_shape(
            current_org_shape, config_name, flow_name, org_shapes
        ) as org_config:
            yield org_config
    else:
        yield None


@contextmanager
def change_org_shape(
    current_org_shape, config_name: str, flow_name: str, org_shapes: dict
):
    # I don't love that we're using the user's real keychain
    # but I get weird popups when I use an empty keychain.
    with unmock_env():
        org_name = f"pytest__{config_name}__{flow_name}"
        org_config = org_shapes.get(org_name)
        if not org_config:
            org_config = _create_org(org_name, config_name, flow_name)
            org_shapes[org_name] = org_config
    current_org_shape.org_config = org_config
    yield org_config


def _create_org(org_name, config_name, flow_name):
    runtime = CliRuntime(load_keychain=True)
    try:
        org, org_config = runtime.get_org(org_name)
    except OrgNotFound:
        org = None
    if org:
        cleanup_org_shapes([org_name])
    org_scratch.callback.__wrapped__(
        runtime,
        config_name,
        org_name,
        default=False,
        devhub=None,
        days=1,
        no_password=False,
    )
    org, org_config = runtime.get_org(org_name)

    coordinator = runtime.get_flow(flow_name)
    coordinator.run(org_config)
    return org_config



from typing import Mapping, Sequence

import sqlalchemy.types as types
from sqlalchemy import (
    Boolean,
    Column,
    ForeignKey,
    Integer,
    PrimaryKeyConstraint,
    String,
    Table,
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.orm.collections import attribute_mapped_collection

Base = declarative_base()


class SpecializedType(types.TypeDecorator):
    """A baseclass for custom datatypes. E.g. sequences and mappings

    Custom types are types that are not directly supported by SQL Alchemy
    """

    impl = types.PickleType

    def process_bind_param(self, value, dialect):
        assert isinstance(value, (self.typ, type(None))), (value, self.typ)
        if not value:
            return None
        return self.empty_value.__class__(value)

    def process_result_value(self, value, dialect):
        if not value:
            return self.empty_value
        return value

    def copy(self, **kw):
        return self.__class__(self.impl.length)


class SequenceType(SpecializedType):
    """Store sequences in an optimized tuple format"""

    typ = Sequence
    empty_value = ()


class MappingType(SpecializedType):
    """Store mappings in an optimized dict format"""

    typ = Mapping
    empty_value = {}


class OrgSchemaModelMixin:
    def __getitem__(self, attr):
        try:
            return getattr(self, attr)
        except AttributeError:
            raise KeyError(attr)

    def __repr__(self):
        fields = self.__dict__.copy()
        name = fields.pop("name", "<UNNAMED>")
        fields.pop("_sa_instance_state")
        return f"<{self.__class__.__name__} {name} {fields}>"

    def __eq__(self, other):
        "Useful mostly for unit tests"
        my_state = self.__dict__.copy()
        my_state.pop("_sa_instance_state")
        other_state = other.__dict__.copy()
        other_state.pop("_sa_instance_state", None)
        return my_state == other_state


class SObject(OrgSchemaModelMixin, Base):
    __tablename__ = "sobjects"
    name = Column(String, primary_key=True, sqlite_on_conflict_primary_key="REPLACE")
    activateable = Column(Boolean)
    childRelationships = Column(SequenceType)
    compactLayoutable = Column(Boolean)
    createable = Column(Boolean)
    custom = Column(Boolean)
    customSetting = Column(Boolean)
    deepCloneable = Column(Boolean)
    defaultImplementation = Column(String)
    deletable = Column(Boolean)
    deprecatedAndHidden = Column(Boolean)
    extendedBy = Column(String)
    extendsInterfaces = Column(String)
    feedEnabled = Column(Boolean)
    fields = relationship(
        "Field",
        collection_class=attribute_mapped_collection("name"),
        back_populates="parent",
    )
    hasSubtypes = Column(Boolean)
    label = Column(String)
    labelPlural = Column(String)
    implementedBy = Column(String)
    implementsInterfaces = Column(String)
    isInterface = Column(Boolean)
    isSubtype = Column(Boolean)
    keyPrefix = Column(String)
    layoutable = Column(Boolean)
    listviewable = Column(String)
    lookupLayoutable = Column(String)
    mergeable = Column(Boolean)
    mruEnabled = Column(Boolean)
    namedLayoutInfos = Column(SequenceType)
    networkScopeFieldName = Column(String)
    queryable = Column(Boolean)
    recordTypeInfos = Column(SequenceType)
    replicateable = Column(Boolean)
    retrieveable = Column(Boolean)
    searchLayoutable = Column(Boolean)
    searchable = Column(Boolean)
    sobjectDescribeOption = Column(String)
    triggerable = Column(Boolean)
    undeletable = Column(Boolean)
    updateable = Column(Boolean)
    urls = Column(MappingType)
    supportedScopes = Column(SequenceType)
    actionOverrides = Column(SequenceType)


field_references = Table(
    "references",
    Base.metadata,
    Column("field_name", Integer, ForeignKey("fields.name")),
    Column("object_name", Integer, ForeignKey("sobjects.name")),
)


class Field(OrgSchemaModelMixin, Base):
    __tablename__ = "fields"
    __table_args__ = (
        PrimaryKeyConstraint(
            "sobject", "name", name="foo", sqlite_on_conflict="REPLACE"
        ),
    )

    sobject = Column(String, ForeignKey("sobjects.name"), nullable=False)
    parent = relationship("SObject", back_populates="fields")
    name = Column(String, nullable=False)
    aggregatable = Column(Boolean)
    aiPredictionField = Column(Boolean)
    autoNumber = Column(Boolean)
    byteLength = Column(Integer)
    calculated = Column(Boolean)
    calculatedFormula = Column(String)
    cascadeDelete = Column(Boolean)
    caseSensitive = Column(Boolean)
    compoundFieldName = Column(String)
    controllerName = Column(String)
    createable = Column(Boolean)
    custom = Column(Boolean)
    defaultValue = Column(String)
    defaultValueFormula = Column(String)
    defaultedOnCreate = Column(Boolean)
    dependentPicklist = Column(Boolean)
    deprecatedAndHidden = Column(Boolean)
    digits = Column(Integer)
    displayLocationInDecimal = Column(Boolean)
    encrypted = Column(Boolean)
    externalId = Column(Boolean)
    extraTypeInfo = Column(types.PickleType)
    filterable = Column(Boolean)
    filteredLookupInfo = Column(types.PickleType)
    formulaTreatNullNumberAsZero = Column(Boolean)
    groupable = Column(Boolean)
    highScaleNumber = Column(Boolean)
    htmlFormatted = Column(Boolean)
    idLookup = Column(Boolean)
    inlineHelpText = Column(String)
    label = Column(String)
    length = Column(Integer)
    mask = Column(String)
    maskType = Column(String)
    name = Column(String)
    nameField = Column(Boolean)
    namePointing = Column(Boolean)
    nillable = Column(Boolean)
    permissionable = Column(Boolean)
    polymorphicForeignKey = Column(Boolean)
    precision = Column(Integer)
    queryByDistance = Column(Boolean)
    referenceTargetField = Column(String)
    referenceTo = Column(types.PickleType)
    relationshipName = Column(String)
    relationshipOrder = Column(Integer)
    restrictedDelete = Column(Boolean)
    restrictedPicklist = Column(Boolean)
    scale = Column(Integer)
    searchPrefilterable = Column(Boolean)
    soapType = Column(String)
    sortable = Column(Boolean)
    type = Column(String)
    unique = Column(Boolean)
    updateable = Column(Boolean)
    writeRequiresMasterRead = Column(Boolean)
    picklistValues = Column(types.PickleType)


class FileMetadata(Base):
    __tablename__ = "file_metadata"
    name = Column(String, primary_key=True, sqlite_on_conflict_primary_key="REPLACE")
    value = Column(String)
